#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>  /*function close(...)*/

#include <Xm/ArrowB.h>    /*motif libraries...*/
#include <Xm/PushB.h>
#include <Xm/Scale.h>
#include <Xm/Text.h>
#include <Xm/ToggleBP.h>
#include <Xm/MwmUtil.h>
#include <Xm/TextF.h>

#include "control.h"
#include "graphics.h"
#include "callback.h"
#include "fitsigma.h"
#include "simu.h"
#include "calc.h"
#include "gvar.h"

/* global variables needed in callback.c */

XtWorkProcId id; /*needed in stopping and restarting the simu*/
XtIntervalId arrowbtimer;
extern Container c[];
extern int  iTschemeMax;

/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/
/*Sets the text-arrow widget taw to sensitivity truth*/
void TASensitive(int taw,int truth)
{
  XtSetSensitive(c[taw].text, truth);
  XtSetSensitive(c[taw].arrow1, truth);
  XtSetSensitive(c[taw].arrow2, truth);
}

/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/
/*
 * text widget value changed callback
 */
void CbTextValueChanged(Widget w, XtPointer clientData, XtPointer callData)
{
  int idn, temp;   
  char buff[6],text[25];
  static int lastavrcutvalue;  /*needed in avrcut text widget*/     
  idn = (int)(long)clientData;  /*get the correct widget*/ 
  *c[idn].value = atof(XmTextGetString(w));
   
  /* check that the new value is in the allowed range*/
  if (*c[idn].value < c[idn].min)      {
    *c[idn].value = c[idn].min;
    sprintf(text, "%-.8g", c[idn].min);
    XmTextSetString(w, text);
  }
  else if (*c[idn].value > c[idn].max) {
    *c[idn].value = c[idn].max;
    sprintf(text, "%-.8g", c[idn].max);
    XmTextSetString(w, text);
  }
  /*a long if to handle different text widgets*/
  if (idn == SIZE_X){
    simArea.x = (1.0 + edge.x / 100.0) * (*c[idn].value);
    /* Simulation cell size has to be adjusted if periodic bc in x */
    adjustsimarea(periodicXFlag,&(simArea.x),lParam);
    /**c[ZOOM_X].value = *c[idn].value + 1;
       sprintf(buff, "%g", *c[ZOOM_X].value); 
       XmTextSetString(c[ZOOM_X].text, buff);*/
  } 
  else if (idn == SIZE_Y){
    simArea.y = (1.0 + edge.y / 100.0) * (*c[idn].value);
    /**c[ZOOM_Y].value = *c[idn].value + 1;
       sprintf(buff, "%g", *c[ZOOM_Y].value);
       XmTextSetString(c[ZOOM_Y].text, buff);*/
  }
  
  if (idn == ZOOM_X || idn == SIZE_X) {
    temp = (int)((simArea.x - zoom.x) / 0.2) + 5;
    if (temp < 0) temp = 0;
    if (center.x > temp / 10) {
      XtVaSetValues(wHScrollB, XmNvalue, temp, NULL);
      center.x = (double)temp / 10.0;
    } 
    else if (center.x < -temp / 10) {
      XtVaSetValues(wHScrollB, XmNvalue, -temp, NULL);
      center.x = -(double)temp / 10.0;
    }
    XtVaSetValues(wHScrollB,
		  XmNminimum,     -temp,
		  XmNmaximum,     temp + 100,
		  NULL);
  } 
  else if (idn == ZOOM_Y || idn == SIZE_Y) {
    temp = (int)((simArea.y - zoom.y) / 0.2) + 5;
    if (temp < 0)
      temp = 0;
    if (center.y > temp / 10) {
      XtVaSetValues(wVScrollB, XmNvalue, temp, NULL);
      center.y = (double)temp / 10.0;
    } 
    else if (center.y < -temp / 10) {
      XtVaSetValues(wVScrollB, XmNvalue, -temp, NULL);
      center.y = -(double)temp / 10.0;
    }
    XtVaSetValues(wVScrollB,
		  XmNminimum,     -temp,
		  XmNmaximum,     temp + 100,
		  NULL);
  }
  /*The following 2 else ifs cause that avrCut can be 0 or >=1 but 
    not for example 0.1 because that would consume huge amounts of memory 
    without a reason.*/
  else if (idn == AVRCUT && *c[idn].value >= 1.0) {
    lastavrcutvalue=avrCut;
    VSCopy(avCells, 1.0 / *c[idn].value, simArea);
    ALLOC(avr, images*avCells.x * avCells.y, VecD);
    ALLOC(avn, images*avCells.x * avCells.y, int);
    ALLOC(avatype,images*avCells.x * avCells.y, int);
    ALLOC(avv, images*avCells.x * avCells.y, VecD);
  }
  else if (idn == AVRCUT && *c[idn].value < 1.0) {
    if (avrCut<=lastavrcutvalue && avrCut>0) {
      avrCut=0;
      XmTextSetString(c[idn].text, "0");
    }
    else if (avrCut>lastavrcutvalue) {
      avrCut=1;
      XmTextSetString(c[idn].text, "1");
    }
    
    lastavrcutvalue=avrCut;
    *c[idn].value=avrCut;
    VSCopy(avCells, avrCut, simArea);  
    ALLOC(avr, images*avCells.x * avCells.y, VecD);
    ALLOC(avn, images*avCells.x * avCells.y, int);
    ALLOC(avatype,images*avCells.x * avCells.y, int);
    ALLOC(avv, images*avCells.x * avCells.y, VecD);
  }
  
  /* 
    Check if we have to build the system anew.
    If you add new text widgets you probably have 
    to change the if clause. See control.h for details.
  */
  if (idn < STRAIN_P) CalcAtoms();
  Redraw();
  if (idn == IMAGESEL && wWinnebFlag) {cgInit=1;RedrawnebPlot();}
  if (idn == CG_LM_STEPS)
    cgMaxLMSteps= (int) cgGraphSteps;
}


/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/


/*
 * text widget modify verification callback
 * this is called when a new character is typed in a text widget
 */
void CbTextModifyVerify(Widget w, XtPointer clientData, XtPointer callData)
{
  int allow,cbnum;
  char *v;
  XmTextVerifyCallbackStruct *cbs = (XmTextVerifyCallbackStruct *)callData;
  cbnum = (long int) clientData;
  switch (c[cbnum].nmdvalue) {
  case TAN:
    allow = ALLOW_NUMBERS;
    break;
  case TAND:
    allow = ALLOW_NUMBERS | ALLOW_DOT;
    break;
  case TANMD:
    allow = ALLOW_NUMBERS | ALLOW_MINUS | ALLOW_DOT;
    break;
  case TANM:
    allow = ALLOW_NUMBERS | ALLOW_MINUS;
    break;
  }
  
  v = ((XmTextVerifyCallbackStruct *)callData)->text->ptr;
  
  /* If the typed character is allowed, return immediately.
     Allowed characters are indicated by different bits of clientData. */
  if (cbs->text->length == 0) {
    return;
  }
  /*else {
    if (cbs->startPos>=c[cbnum].length) {
      cbs->doit = False;
      return;
    }
  }*/
  if (v == NULL) return;
  else if (v[0] == 0) /* back space is always allowed */
    return;
  else if ((allow & ALLOW_NUMBERS) && v[0] >= '0' && v[0] <= '9')
    return;
  else if ((allow & ALLOW_MINUS) && v[0] == '-')
    return;
  else if ((allow & ALLOW_DOT) && v[0] == '.')
    return;
  else
    ((XmTextVerifyCallbackStruct *)callData)->doit = False;
}


/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * text widget lose focus callback
 */
void CbTextLoseFocus(Widget w, XtPointer clientData, XtPointer callData)
{
  
  double temp;
  char buff[6];
  int id;

  id = (int)(long)clientData;
  temp = atof(XmTextGetString(w));
  if (temp != *c[id].value) {
    sprintf(buff, "%g", *c[id].value);
    XmTextSetString(w, buff);
  }
}


/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*arrowbutton widget's callback. This handles holding down the button which
  causes the values to change without clicking. This calls CbArrow which
  makes the main job. All this procedure does is to stop changing the value
  when the button disarms and to call cbarrow when the button arms*/

void CbArrowTimerStart(Widget wid,XtPointer client_data,XtPointer call_data)
{
  XmArrowButtonCallbackStruct *cbs = (XmArrowButtonCallbackStruct *)
    call_data;
  if (cbs->reason == XmCR_ARM)
    CbArrow(client_data,1);
  else if (cbs->reason == XmCR_DISARM)
    if (arrowbtimer != 0)
      XtRemoveTimeOut(arrowbtimer);
}


/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * arrowbutton widget's callback. CbTextValueChanged is the main callback
 * this doesn't have that much to do, just update the value if that is
 * allowed. This is called only when a textwidget is updated by pressing
 * the arrows. CbTextValueChanged is called always when the text changes,
 * whatever the cause. This procedure also starts a timer which handles
 * changing the textarrowwidget's value constantly when pressing the mouse
 * button.
 */
void CbArrow(XtPointer clientData, XtIntervalId timerid)
{
  char buff[6];
  double temp;
  int id;
  
  id = (int)(long)clientData / 2;
  /*a switch case to check all necessary textwidgets.*/ 
  switch (id) {
  case ROT_R:
  case BND_D:
  case RELAXATION:
  case DISP_X:
  case DISP_Y: 
    if (!boundaryFlag) return;
    break;    /*these widgets not in use if no boundary*/

  case CRACK_X:
  case CRACK_Y:
  case CRACK_SX:
  case CRACK_SY:
  case CRACK_ROT: 
    if (!crackFlag) return;
    break;    /*these widgets not in use without crack*/

  case COOLTEMP:
  case SLOPE:
  case STRTTIME: 
    break;     

  case DOT:
    CalcAtoms();
  case UPCUTA:
  case DOWNCUTA: 
  case UPCUT:
  case DOWNCUT:
    break;

  case WAIT:     
    break;

  case AVRCUT:  
    if (sinfo != SVECT) return;
    break;

  case BEGINAT:
  case ENDAT:   
    if (!ani.limitsFlag) return;
    break;
  }

  /*check that the new value is in the allowed range*/
  temp = *c[id].value + (((long)clientData%2==0L)?-1.0:1.0)*c[id].step;
  if (temp < c[id].min) {
    temp = c[id].min;
    return;
  }
  else if (temp > c[id].max) {
    temp = c[id].max;
    return;
  }
  if (temp != *c[id].value) {
    sprintf(buff, "%g", temp);
    XmTextSetString(c[id].text, buff);   /*update the text to max/min value*/
    /*start the timer. On the first time (timerid==1) wait longer before
      repeating the callback*/
    arrowbtimer=XtAppAddTimeOut(app,timerid==1? 500 : 50,
				(XtTimerCallbackProc)CbArrow,clientData); 
  }
}



/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * Button callback - both pushb and toggleb. This is called always when a
 * sensitive button is pressed.
 */
void CbButton(Widget w, XtPointer clientData, XtPointer callData)
{
  char buff1[MAXLINE],buff2[MAXLINE];
  /* a switch case for all buttons that do something*/
  int image,atom,atom0,atom1,atom2;
  /*double maxI,minI;*/
  int minI,maxI,i;
  switch ((long)clientData) {
 
  case BOUNDARY: 
    boundaryFlag = !boundaryFlag;
    /*XtSetSensitive(c[ROT_R].text, boundaryFlag);
    XtSetSensitive(c[BND_D].text, boundaryFlag);*/
    CalcAtoms();
    Redraw();
    break;
  
  case VERTIC:   
    verticalFlag= !verticalFlag;
    if (verticalFlag && periodicXFlag) {
      XtVaSetValues(wPerxToggle, XmNset,False, NULL);
      XtCallCallbacks(wPerxToggle,XmNvalueChangedCallback,NULL);
    }  
    
    XtSetSensitive(wPerxToggle, !verticalFlag);
        
    CalcAtoms();
    Redraw();
    break;
    
  case SUGGEST:   
    suggestFlag= !suggestFlag;
    break;

  case PERX:
    periodicXFlag =! periodicXFlag;
    /*if (periodicXFlag) 
      edge.x=0.0;
    else
      edge.x=wDrawSize.x;
    simArea.x  = (1.0 + edge.x / 100.0) * size.x;
    simArea.y  = (1.0 + edge.y / 100.0) * size.y;
    adjustsimarea(periodicXFlag,&(simArea.x),lParam);*/
    CalcSimArea();
    CalcAtoms();
    Redraw();
    break;

  case CRACK: 
    crackFlag = !crackFlag;
    TASensitive(CRACK_X, crackFlag); 
    TASensitive(CRACK_Y, crackFlag); 
    TASensitive(CRACK_SX, crackFlag); 
    TASensitive(CRACK_SY,  crackFlag); 
    TASensitive(CRACK_ROT,  crackFlag); 
    CalcAtoms();
    Redraw();
    break;

  case TAPPLY:
    tapplyFlag = !tapplyFlag;
    break;
  case TADJUST:
    tadjustFlag = !tadjustFlag;
    break; 
  case TADJUST2:
    tadjust2Flag = !tadjust2Flag;
    break; 
  case TSAVE:
    savetscheme = !savetscheme;
    break; 
  
  case RUN: 
    if (runningFlag)
      XtRemoveWorkProc(id);
    else
      id = XtAppAddWorkProc(app, MdCycle, 0);
    runningFlag = !runningFlag; /* toggle run state */
    XtVaSetValues(wRunPushB, XmNalignment,XmALIGNMENT_CENTER,XtVaTypedArg,
		  XmNlabelString, T_STRING(runningFlag ? "Pause" : "Run"),
		  NULL);
    break;
    
  case PARAM_OK: 
    cImage=0;
    grImage=0;
    if (wWinnebFlag) XtPopdown(XtParent(wnebMan));
    XtPopdown(XtParent(wParamMan));
    if(wWinSeamFlag) XtPopdown(XtParent(wSeamMan));
    strainFlag = False;
    CalcAtoms();
    InitRun();
    XtSetSensitive(wRunPushB, True);
    tapplyFlag = False;
    iTschemeMax=-1;
    if (tapplyFlag) {
      readTscheme(tfname); 
      checkTscheme();
    }
    if (savetscheme) {
      if (! (tsavefile = fopen ("save.events", "w"))){
	fprintf(stderr,"Could not open file save.events\n");
	bexit(-1);
      }
      fprintf(tsavefile,"#\n#Auto-saved partial event list\n#\n");
      fprintf(tsavefile,"Size %.2f %.2f\n\n",size.x,size.y);
    }
    Redraw();
    //RedrawPlot(1);
    break;

  case CANCEL:   
    XtPopdown(XtParent(wParamMan));
    if(wWinSeamFlag) XtPopdown(XtParent(wSeamMan));
    XtSetSensitive(wRunPushB, True);
    SaveParams(LOAD);
    Redraw();
    break; 

  case SEAM_OK: 
    XtPopdown(XtParent(wSeamMan));
    XtSetSensitive(wSeamPushB, True);
    wWinSeamFlag=0;
    break;
    
  case SEAM:  
    wWinSeamFlag=1;
    /*XtSetSensitive(wSeamPushB, False);*/
    OpenDialog(wSeamMan,w,"seam");
    break;

    /*-----------------------------------------------*/    

  case EVENTS_OK:
    readTscheme(tfname); 
    tapplyFlag = True;
    checkTscheme();
    XtPopdown(XtParent(wEventsMan));
    Redraw();
    break;
  case EVENTS_CANCEL:
    XtPopdown(XtParent(wEventsMan));
    break;
    /*-----------------------------------------------*/

  case STRAIN: 
    strainStep = 0.005 * strainP * (xMax - xMin) / 
      (strainT / (1000 * deltaT));
    XtSetSensitive(wStrainPushB, False);
    XtSetSensitive(wFixedToggle, False);
    TASensitive(STRAIN_T, False);
    TASensitive(STRAIN_P, False);
    strainFlag = True;
    break;
    
  case FIXED_TOGGLE: 
    fixedFlag = !fixedFlag;
    XtSetSensitive(wStrainPushB, fixedFlag && runningFlag);
    Redraw();
    break;

  case CLOSE_STRAIN:   
    XtPopdown(XtParent(wStrainMan));
    /*Redraw();*/
    break;

/*-----------------------------------------------*/

  case TSCALE_TOGGLE: 
    cool.TScaleFlag = !cool.TScaleFlag;
    XtVaSetValues(wTScaleToggle,XmNset,cool.TScaleFlag,NULL); 
    
    /* write to save.events */
    if (savetscheme && cool.TScaleFlag) 
      fprintf(tsavefile,
	      "TScaleOn %.2f %f\n",
	      (stepCount * 1000 *deltaT), cool.tempvalue); 
    else if (savetscheme && !cool.TScaleFlag)
      fprintf(tsavefile,
	      "TScaleOff %.2f\n",
	      (stepCount * 1000 *deltaT)); 
    break;
    
  case QUENCH_TOGGLE: 
    cool.QuenchFlag = !cool.QuenchFlag;
    XtVaSetValues(wQuenchToggle,XmNset,cool.QuenchFlag,NULL); 
    
    /* write to save.events */
    if (savetscheme && cool.QuenchFlag) 
      fprintf(tsavefile,
	      "QuenchOn %.2f\n",
	      (stepCount * 1000 *deltaT)); 
    else if (savetscheme && !cool.QuenchFlag)
      fprintf(tsavefile,
	      "QuenchOff %.2f\n",
	      (stepCount * 1000 *deltaT)); 
    break;
  case ALTMIN: 
    cgFlag = !cgFlag;
    cgInit=1;
    if (!cgFlag) {
      for (i=cImage*nAtom; i < (cImage+1)*nAtom; i++) {
	ae[i].rv.x=0.0;
	ae[i].rv.y=0.0;
      }
    }
    break;
  case CG_SEMILM:
    cgSemiLMFlag = !cgSemiLMFlag;
    break; 
  case CG_LINEMIN: 
    cgLineMinFlag = !cgLineMinFlag;
    break;
  case SEMISTEEP: 
    semisteepFlag = !semisteepFlag;
    cgInit=1;
    break;
  case TEMP_LANGEVIN: 
    cool.langevinFlag = !cool.langevinFlag;
    break;
  case RANDOMIZE:
    RandomizeVelocities();

    /* write to save.events */
    if (savetscheme) 
      fprintf(tsavefile,
	      "Randomize %.2f\n",
	      (stepCount * 1000 *deltaT)); 
    break;
  
  case START_COOL: 
    InitRun();
    cool.programon =1;
    strainFlag = False;
    Redraw();
    //RedrawPlot(1);
    if (runningFlag)
      XtRemoveWorkProc(id);
    else
      id = XtAppAddWorkProc(app, MdCycle, 0);
    runningFlag = !runningFlag; /* toggle run state */
    /*XtSetSensitive(wStrainPushB,
      runningFlag && !strainFlag && fixedFlag);*/
    XtSetSensitive(wCoolPushB,False);
    TASensitive(COOLTEMP, False);
    TASensitive(STRTTIME, False);
    TASensitive(SLOPE, False);
    
    XtVaSetValues(wRunPushB,  XmNalignment,XmALIGNMENT_CENTER,
		  XtVaTypedArg,
		  XmNlabelString, T_STRING(runningFlag ? "Pause" : "Run"),
		  NULL);
    Redraw();
    break;
    
  case CLOSE_TEMP:   
    XtPopdown(XtParent(wTempMan));
    break;
    
    /*-----------------------------------------------*/
  case MODIFY:
    ModifyArea(modarea);
    if (savetscheme) {
      switch (modinfo) {
      case M_MODRM:
	sprintf(buff1, "remove");
	break;
      case M_MODSTOP:
	sprintf(buff1, "stop");
	break;
      case M_MODTYPE1:
	sprintf(buff1, "type1");
	break;
      case M_MODTYPE2:
	sprintf(buff1, "type2");
	break;	
      case M_MODVEL:
	sprintf(buff1, "vel");
	break;
      case M_MODFIX:
	sprintf(buff1,"fix");
	break;
      case M_MODFIXX:
	sprintf(buff1,"fixx");
	break;
      case M_MODFIXY:
	sprintf(buff1,"fixy");
	break;
      case M_MODFORCE:
	sprintf(buff1, "force");
	break;
      case M_MODRMF:
	sprintf(buff1,"rmfix");
	break;
      case M_MODMOVE:
	sprintf(buff1,"move");
	break;
      case M_MODTAG:
	sprintf(buff1,"tag");
	break;
      case M_MODROT:
	sprintf(buff1,"rot");
	break;
      }	

      fprintf(tsavefile,
	      "ModifyArea %.2f %s  %.2f %.2f  %.2f %.2f  %.2f %.2f  %.2f\n",
	      (stepCount * 1000 *deltaT), buff1, modvel.x, modvel.y, 
	      modarea.center.x, modarea.center.y,
	      modarea.size.x,modarea.size.y,modarea.angle); 
    }  
    ComputeForces();
    Redraw();
    /*}*/
    break;

  case CLOSE_MOD:   
   XtPopdown(XtParent(wModMan));
   wWinModFlag = False;
   break;
  case MODALL:
    modallFlag = !modallFlag;
    break;
  case MOD_RELAT:
    modrelatFlag = !modrelatFlag;
    break;
  case MODSHOWSEL:
    showselFlag = !showselFlag;
    Redraw();
    break;
    
    /*-----------------------------------------------*/
  case NEB_COPY:
    if (nebcpi != grImage) {
      for (atom=0; atom<nAtom; atom++) {
	atom1=grImage*nAtom+atom;
	atom2=nebcpi*nAtom+atom;
	ae[atom2]=ae[atom1];
      }
    }
    
    neb[gtype(nebcpi)].pot=neb[gtype(grImage)].pot;
    neb[gtype(nebcpi)].pot = neb[gtype(grImage)].pot;
    nebcpi++;
    if (nebcpi > (images-1)) nebcpi=0;
    sprintf(buff1,"%d",(int)nebcpi);
    XmTextSetString(c[IMAGECP].text, buff1);
    break;
  case NEB_IP:
    /* neb initial interpolation */
    Interpolate(nebip1,nebip2);
    break;

    /* add neb images */
  case NEB_ADDI:
    if ((nebaddamount > 0) && (grImage < images-1)) 
      AddImages(grImage,nebaddamount);
    break;

  case NEB_FIX:
    if (nebip2 > nebip1) {
      maxI = (int) nebip2;
      minI = (int) nebip1;
    } else {
      maxI = (int) nebip1;
      minI = (int) nebip2;
    }
    for (image=minI; image < maxI+1; image++) {
      neb[image].fix=1;
    }
    break;

  case NEB_RMFIX:
    if (nebip2 > nebip1) {
      maxI = (int) nebip2;
      minI = (int) nebip1;
    } else {
      maxI = (int) nebip1;
      minI = (int) nebip2;
    }
    for (image=minI; image < maxI+1; image++) {
      neb[image].fix=0;
    }
    break;

  case NEB_CSTOP:
    nebcStop = !nebcStop;
    break;

  case NEB_RELAX:
    nebrelFlag = !nebrelFlag;
    break;

  case NEB_REP:
    nebrepFlag = !nebrepFlag;
    break;

    /*-----------------------------------------------*/
  case CLOSE_ZOOM:   
    XtPopdown(XtParent(wZoomMan));
    break;

    /*-----------------------------------------------*/

  case CLOSE_COLOR:   
    XtPopdown(XtParent(wColorMan));
    break;

  case COLRANUPD:   
    colorranges.Update = !colorranges.Update;
    if (!colorranges.Update) {
      sprintf(buff1, "%.2g", colorranges.EkMin);
      sprintf(buff2, "%.2g", colorranges.EkMax);
      XmTextSetString(c[COLRANKMIN].text, buff1);
      XmTextSetString(c[COLRANKMAX].text, buff2);
      
      sprintf(buff1, "%.2g", colorranges.EpMin);
      sprintf(buff2, "%.2g", colorranges.EpMax);
      XmTextSetString(c[COLRANPMIN].text, buff1);
      XmTextSetString(c[COLRANPMAX].text, buff2);
      
      sprintf(buff1, "%.2g", colorranges.SMin);
      sprintf(buff2, "%.2g", colorranges.SMax);
      XmTextSetString(c[COLRANSMIN].text, buff1);
      XmTextSetString(c[COLRANSMAX].text, buff2);

      sprintf(buff1, "%.2g", colorranges.FMin);
      sprintf(buff2, "%.2g", colorranges.FMax);
      XmTextSetString(c[COLRANFMIN].text, buff1);
      XmTextSetString(c[COLRANFMAX].text, buff2);
    }
    TASensitive(COLRANKMIN, !colorranges.Update);
    TASensitive(COLRANKMAX, !colorranges.Update);
    TASensitive(COLRANPMIN, !colorranges.Update);
    TASensitive(COLRANPMAX, !colorranges.Update);
    TASensitive(COLRANSMIN, !colorranges.Update);
    TASensitive(COLRANSMAX, !colorranges.Update);
    TASensitive(COLRANFMIN, !colorranges.Update);
    TASensitive(COLRANFMAX, !colorranges.Update);
    break;

  case COLRANHIDE:   
    colorranges.hide = !colorranges.hide;
    Redraw();
    break;
    /*-----------------------------------------------*/

  case SETTRACKTOG: 
    SetTrackFlag = !SetTrackFlag; 
    TASensitive(TRANGE, SetTrackFlag);
    TASensitive(TFACTOR, SetTrackFlag);
    TASensitive(TBFACTOR, SetTrackFlag);
    XtSetSensitive(wSaveTrackToggle, SetTrackFlag);   
    XtSetSensitive(wShowTrackToggle, SetTrackFlag);   
    break;

  case SAVETRK:
    saveTrackFlag=!saveTrackFlag;
    if (saveTrackFlag){
      FileFlag=True;
    }
    else 
      if (FileFlag){
	StartFlag=False;
      }
    break;  
    
  case SHOWTRK:
    showTrackFlag=!showTrackFlag;
    break;  
    
  case CLOSE_TRK: 
    XtPopdown(XtParent(wTrackMan));
    wWinTrackFlag=False;
    break;

/*-----------------------------------------------*/
  
  case SET_ANI: 
    ani.aniFlag = !ani.aniFlag;
    break;
        
  case CLOSE_ANI: 
    XtPopdown(XtParent(wAniMan));
    wWinAniFlag = False;
    break;
 
  case LIM_TOGGLE: 
    ani.limitsFlag = !ani.limitsFlag; 
    TASensitive(BEGINAT, ani.limitsFlag);
    TASensitive(ENDAT, ani.limitsFlag);
    XtSetSensitive(wEndendToggle, ani.limitsFlag);
    break;
    
  case ENDEND: 
    ani.endendFlag = !ani.endendFlag;
    break; 
  
  }
}
/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

void Interpolate(int image1,int image2)
{
  int maxI,minI;
  int image,atom,atom0,atom1,atom2;
  double ddx;

  if (image2 > image1) {
    maxI = image2;
    minI = image1;
  } else {
    maxI = image1;
    minI = image2;
  }


  if (maxI > minI+1) {
    for (image=minI+1; image<maxI; image++) {
      for (atom=0; atom<nAtom; atom++) {
	atom0 = minI*nAtom+atom;
	atom1 =  image*nAtom+atom;
	atom2 = maxI*nAtom+atom;
	
	ddx=ae[atom2].r.x - ae[atom0].r.x;
	if (periodicXFlag) PBCdx(ddx);
	ae[atom1].r.x = ae[atom0].r.x + ddx/(maxI-minI)*((double)(image-minI));
	if (periodicXFlag) PBCx(ae[atom1].r.x);
	ae[atom1].r.y = ae[atom0].r.y + (ae[atom2].r.y - ae[atom0].r.y)
	  /(maxI-minI)*((double)(image-minI));
      }
    }
  }
}



/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/


AtomD* allocad(AtomD *array,int amount)
{
  AtomD *tmp;
  int i;

  if ((tmp = (AtomD*)malloc(images * (nAtom) * sizeof(AtomD))) == NULL) {
    printf("Error allocating memory for tmp!\n");
    bexit(-1);
  }
  
  for (i=0; i < nAtom*images; i++) {
    tmp[i]=array[i];
  }
  ALLOC(array, (images+amount) * nAtom, AtomD);
 
  for (i=0; i < nAtom*images; i++) {
    array[i]=tmp[i];
  }
  free(tmp);
  return array;
}


/* adds neb images */
void AddImages(int min, int amount)
{
  int i,a;

  ae=allocad(ae,amount);

  ALLOC(neb,(images+amount), nebImage);


  for (i=images-1; i > min; i--) {
    for (a=0; a < nAtom; a++) {
      ae[(i+amount)*nAtom+a]       = ae[i*nAtom+a];
    }
    neb[i+amount] = neb[i];
      
  }
  for (i=min+1; i < min+amount+1; i++) {
    for (a=0; a < nAtom; a++) {
      ae[i*nAtom+a]       = ae[min*nAtom+a];
    }
    
    neb[i]=neb[min];
    neb[i].fix = 0;
  }

  images += amount;
  Interpolate(min,min+amount+1);


  c[IMAGESEL].max = images-1;
  c[NEB_IP1].max = images-1;
  c[NEB_IP2].max = images-1;
  c[IMAGECP].max = images-1;
} 


/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/
/*returns 1/true if atom i is in the area, else 0/false*/
int Selected(long int i, RArea area)
{
  double dx,dy,angle;

  angle=area.angle*Pi/180.0;
  dx=(ae[i].r.x-area.center.x)*cos(angle)+(ae[i].r.y-area.center.y)*sin(angle);
  dy=(ae[i].r.y-area.center.y)*cos(angle)-(ae[i].r.x-area.center.x)*sin(angle);
  if ((fabs(dx) <= 0.5*area.size.x) 
      && (fabs(dy) <= 0.5*area.size.y))
    return 1;
  else
    return 0;
}

/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/* Removes one atom */
void RemoveAtom(int i) {
  int j;
  if (modallFlag) {
    for (j=0;j<images;j++) {
      ae[j*nAtom+i] =     ae[(j+1)*nAtom-1];
    }
  } else {
    ae[((int)grImage)*nAtom+i]=ae[((int)grImage+1)*nAtom-1];
  }
  nAtom--;
}

/* Modifies atoms in a given area */
void ModifyArea(RArea area)
{
  int i,j,k,min,max;
  double dx,dy,xx,yy,phi,cphi,sphi;
  
  if (modallFlag && wWinnebFlag) {
    min = -((int)grImage)*nAtom;
    max = (images-((int)grImage))*nAtom;
  } else {
    min=0;max=1;
  }
  if (modinfo != M_MODRM) {
    for (i = nAtom*grImage; i < nAtom*(grImage+1); i++) {
      
      if (Selected(i,area)) {
	for (j=min; j<max; j+=nAtom) {
	  switch (modinfo) {
	  case M_MODSTOP:
	    ae[i+j].rv.x = 0.0;
	    ae[i+j].rv.y = 0.0;
	    break;
	  case M_MODTYPE1:
	    ae[i+j].atype = 1;
	    break;
	  case M_MODTYPE2:
	    ae[i+j].atype = 2;
	    break;
	  case M_MODVEL:
	    ae[i+j].rv.x = ae[i+j].rv.x + modvel.x;
	    ae[i+j].rv.y = ae[i+j].rv.y + modvel.y;
	    break;
	  case M_MODFIX:
	    ae[i+j].fixvel = FIXX | FIXY;
	    ae[i+j].rv.x = modvel.x;
	    ae[i+j].rv.y = modvel.y;
	    break;
	  case M_MODFIXX:
	    ae[i+j].fixvel = FIXX;
	    ae[i+j].rv.x = modvel.x;
	    break;
	  case M_MODFIXY:
	    ae[i+j].fixvel= FIXY;
	    ae[i+j].rv.y = modvel.y;
	    break;
	  case M_MODRMF:
	    ae[i+j].fixvel=0;
	    ae[i+j].cforce.x = 0.0;
	    ae[i+j].cforce.y = 0.0;
	    break;
	  case M_MODFORCE:
	    ae[i+j].cforce.x = modvel.x;
	    ae[i+j].cforce.y = modvel.y;
	    break;
	  case M_MODMOVE:
	    ae[i+j].r.x += modvel.x;
	    ae[i+j].r.y += modvel.y;
	    break;
	  case M_MODTAG:
	    ae[i+j].atag = gtype(modvel.x*10.0);
	    break;
	  case M_MODROT:
	    if (!wWinnebFlag) {
	      cphi=cos(modvel.x/57.29577951);
	      sphi=sin(modvel.x/57.29577951);
	      xx=ae[i+j].r.x-area.center.x;
	      yy=ae[i+j].r.y-area.center.y;
	      ae[i+j].r.x= xx*cphi+yy*sphi+area.center.x;
	      ae[i+j].r.y=-xx*sphi+yy*cphi+area.center.y;
	    }
	    break;
	  }
	}
      }
    }       
  } else { /* We have M_MODRM */
    if (wWinnebFlag) {
      for (i=0; i<nAtom; i++) {
	if (Selected(i+nAtom*grImage,area)) {
	  for (j=0; j < images; j++) {
	    ae[j*nAtom+i] = ae[(j+1)*nAtom-1];
	  }
	  for (j=1; j < images; j++) {
	    for (k=0; k < nAtom; k++) {
	      ae[j*nAtom+k-j]=ae[j*nAtom+k];
	    }
	  }
	  i--;
	  nAtom--;
	  //fprintf(stdout,"nAtom: %d\n",nAtom);
	}
      }
    } else {
      for (i=grImage*nAtom; i < nAtom*(grImage+1); i++) {
	if (Selected(i, area)) {
	  ae[i] = ae[((int)grImage)*nAtom+nAtom-1];
	  i--;
	  nAtom--;
	}
      }
    }
  }

  if (modinfo==M_MODROT) RmOverlappingAtoms(grImage,Rminol,area);

}

/*---------------------------------------------------------*/
/*---------------------------------------------------------*/
/* Writes MEP plot data to disk. */
void SaveMEP(int fnum) {
  FILE *mepfile;
  int i;
  char filename[12];
  double tkink;
  sprintf(filename,"%smep.%04d",savedir,fnum);
  if (! (mepfile = fopen (filename, "w"))){
    fprintf(stderr,"Could not open file %s\n",filename);
    bexit(-1);
  }
  fprintf(mepfile,"### Saved MEP format: <image> <pot.energy> <1-dim. coord>"
	  " <cos(path angle)>\n");
  for (i=0; i < images; i++) {
    if (i==0||i==images-1) {
      tkink=1.0;
    } else {
      tkink=neb[i].kink;
    }
    fprintf(mepfile,"%d %.10g %.10g %.10g\n",i,neb[i].pot, neb[i].x, tkink);
  }
  fclose(mepfile);
  fprintf(stdout,"Wrote %s\n",filename);
}

/*---------------------------------------------------------*/
/*---------------------------------------------------------*/

/* Saves atom positions to disk. This is almost directly called when the user
   selects "Save snapshot" from the file menu. 
   Flags PDBFORMAT, DUMPWINDOW, STOREVELOCITIES, TFIELDSCALING are defined
   in callback.h
*/
void SnapShot()
{
  Window dumpId;
  static char cmd_string[200];
  static int filenum = 0;  
  double tfield;
  int i,imag;
  FILE *cd;
  static char bitfile_name[15],paramfile_name[25],line[MAXLINE],temp[4];
  static char aname[5][3];
  int ity;
  int maximage;
  char temps[MAXLINE];
  int length;
  /* 
               atom      atom 
     atype[]   type      name    
     -------  --------   ----
       1      1 free      SI
       2      2 free      GE
      -1      1 fixed     GA
      -2      2 fixed     AS

  */
  
  strcpy(aname[0],"AS\0");
  strcpy(aname[1],"GA\0");
  strcpy(aname[2],"??\0");
  strcpy(aname[3],"SI\0");
  strcpy(aname[4],"GE\0");
  filenum += 1;
  
  /*checks if there are earlier snapshots and doesn't overwrite them*/
  
  //system("ls coords.*.i*.x?? 2>/dev/null | sort -n -r 1>tmxfllst.txt");
  if (strlen(savedir) > 0) {
    sprintf(temps,"mkdir %s",savedir);
    system(temps);
  }
  sprintf(temps,
	  "ls %scoords.*.i*.x?? 2>/dev/null | sort -n -r 1>tmxfllst.txt",
	  savedir);
  system(temps);
  cd=fopen("tmxfllst.txt","r");
  //fprintf(stderr,"\n%s\n",temps);
 
  //length=strlen(line);
  //fprintf(stderr,"%d",length);bexit(0);

  if (fgets(line,MAXLINE,cd) != NULL)  {
    length=strlen(line);//fprintf(stderr,"\n\n%d\n\n",length);bexit(0);
    temp[0]=line[length-14];temp[1]=line[length-13];
    temp[2]=line[length-12];temp[3]=line[length-11];
    //fprintf(stderr,"\n\n%c\n\n",line[length-11]);bexit(0);
    //temp[0]=line[7];temp[1]=line[8];temp[2]=line[9];temp[3]=line[10];
    filenum=atol(temp)+1;
  }
  fclose(cd);
  system("rm tmxfllst.txt 2>/dev/null");
  //fprintf(stderr,"\n\n%s\n\n",temp);bexit(0);
  /*save boundary params with the snapshot so that a situation can be reloaded
   later without changing the parameters*/
  sprintf(paramfile_name,"%sboundary.%04d.param",savedir, filenum);
  ROWParams("write",paramfile_name);

  //bexit(-1);

  if (wWinnebFlag)
    maximage = images;
  else
    maximage = 1;
  
  for (imag=0; imag < maximage; imag++) {

    if (PDBFORMAT) sprintf(bitfile_name,"%scoords.%04d.i%03d.pdb",savedir,
			   filenum,imag);
    else sprintf(bitfile_name,"%scoords.%04d.i%03d.xyz",savedir,
		 filenum,imag);
    if (! (cd = fopen (bitfile_name, "w"))){
      fprintf(stderr,"Could not open file %s: ",bitfile_name);
      perror("");
      fprintf(stderr,"Exiting.\n");
      bexit(-1);
    }
    
    fprintf(stderr,"%d %s %ld %g\n",filenum,bitfile_name,
	    stepCount,deltaT*stepCount);
    
   


    if (PDBFORMAT) { /* Dump in PDB format with potEn as 'temperature' field */
      fprintf(cd,"COMPND    Frame number %ld %g\n",stepCount,deltaT*stepCount);
      for(i=imag*nAtom;i<nAtom*(imag+1);i++) {
	if (sinfo == SPOT) 
	  tfield=TFIELDSCALING*ae[i].potEn;
	else
	  tfield=TFIELDSCALING*ae[i].stressij;    
	ity=ae[i].atype+2; 
	fprintf(cd,"ATOM  %5d%3s   MOL     1    %8.3f%8.3f%8.3f%6.2f%6.2f\n",
		i+1,aname[ity],ae[i].r.x,ae[i].r.y,0.0,1.0,tfield);
      }
      fprintf(cd,"TER   %5d\n",nAtom+1);
      fprintf(cd,"END\n");
    } 
    else { /* Dump XYZ in format */
      fprintf(cd,"%d\n",nAtom);
      fprintf(cd,"Frame number %ld %g\n",stepCount,deltaT*stepCount);
      
      for(i=imag*nAtom;i<nAtom*(imag+1);i++){
	if (sinfo == SPOT) 
	  tfield=TFIELDSCALING*ae[i].potEn;
	else
	  tfield=TFIELDSCALING*ae[i].stressij;    
	ity=ae[i].atype+2; 
	fprintf(cd,"%s %f %f %f %g\n",aname[ity],ae[i].r.x,ae[i].r.y,
		0.0,tfield); 
      }
    }
    
    fclose(cd);
    
    /* save extra atom info: fixvel, cforce etc */
    sprintf(bitfile_name,"%sadata.%04d.i%03d.dat",savedir,filenum,imag);
    if (cd = fopen (bitfile_name, "w")) {
      fprintf(cd,"###Extra atom info, format: <atom> <fixvel> <cforce.x>"
	      " <cforce.y> <atag>\n");
      for(i=imag*nAtom;i<nAtom*(imag+1);i++) {
	if (ae[i].fixvel || ae[i].cforce.x || ae[i].cforce.y || ae[i].atag)
	  fprintf(cd,"%d %d %g %g %d\n", i, ae[i].fixvel, ae[i].cforce.x,
		  ae[i].cforce.y, ae[i].atag);
      }
      fclose(cd);
    } else {
      fprintf(stderr,"Could not open file %s: ",bitfile_name);
      perror("");
      fprintf(stderr,"Exiting.\n");
      bexit(-1);
    } 
   
    
    /*-----end of save extra info----------------------*/ 

    if (STOREVELOCITIES) {
      sprintf(bitfile_name,"%svelocities.%04d.i%03d.dat",savedir,
	      filenum,imag);
      if (! (cd = fopen (bitfile_name, "w"))){
	fprintf(stderr,"Could not open file %s: ",bitfile_name);
	perror("");
	fprintf(stderr,"Exiting.\n");
	bexit(-1);
      } 
      for(i=imag*nAtom;i<nAtom*(imag+1);i++) 
	fprintf(cd,"%d %g %g %g %g\n",i,ae[i].r.x,ae[i].r.y,ae[i].rv.x,
		ae[i].rv.y);
      fclose(cd);
    }
  }
  if (DUMPWINDOW) { /* Dump drawing area to a xwd file */
    dumpId=XtWindow(wDraw);
    sprintf(cmd_string, "(sleep 5; xwd -frame -out %sdump.%04d.xwd -id 0x%x)&",
	    savedir, filenum,(unsigned int)dumpId);
    system(cmd_string);
  }
  if (wWinnebFlag) SaveMEP(filenum);
}



/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * temporarily save and load initial parameters. Used by OK and cancel 
 * buttons in the parameters dialog.
 */
void SaveParams(int function)
{
  /*all parameters that are saved and loaded.*/
  static VecD sizeTemp, simAreaTemp;
  static RArea crackTemp;
  static Seamdata seamTemp;
  static double lParamTemp,lParam2Temp;
  static double cutTemp, borderWTemp,epsilonTemp;
  static int boundaryFlagTemp, suggestFlagTemp,periodicXFlagTemp;
  /*static VecD *rTemp;
  static int *atypeTemp;
  static int *borderatomTemp;*/
  static AtomD *aeTemp;
  static int nAtomTemp;
  static int savetschemeTemp,tapplyFlagTemp,tadjustFlagTemp;
  
  if (function == STORE) {
    sizeTemp.x = size.x;
    sizeTemp.y = size.y;
    simAreaTemp.x = simArea.x;
    simAreaTemp.y = simArea.y;
    seamTemp = seam;
    crackTemp = crack;
    lParamTemp = lParam;
    lParam2Temp = lParam2;
    epsilonTemp = epsilon;
    cutTemp = cut;
    borderWTemp = borderW;
    boundaryFlagTemp = boundaryFlag;
    suggestFlagTemp = suggestFlag;
    periodicXFlagTemp = periodicXFlag;
    savetschemeTemp=savetscheme;
    tapplyFlagTemp=tapplyFlag;
    tadjustFlagTemp=tadjustFlag;
    aeTemp = ae;
    ae = NULL;
    nAtomTemp = nAtom;
  }
  else {
    free(ae);
    ae = aeTemp;
    nAtom = nAtomTemp;
    size.x = sizeTemp.x;
    size.y = sizeTemp.y;
    simArea.x = simAreaTemp.x;
    simArea.y = simAreaTemp.y;
    seam=seamTemp;
    crack = crackTemp;
    lParam = lParamTemp;
    lParam2 = lParam2Temp;
    epsilon = epsilonTemp;
    cut = cutTemp;
    borderW = borderWTemp;
    boundaryFlag = boundaryFlagTemp;
    periodicXFlag = periodicXFlagTemp;
    suggestFlag=suggestFlagTemp;
    savetscheme=savetschemeTemp;
    tapplyFlag=tapplyFlagTemp;
    tadjustFlag=tadjustFlagTemp;
  }
}


/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * scale widget's value-changed callback. Only 3 scales in the program
 * currently so this is a short one.
 */
void CbScale(Widget w, XtPointer clientData, XtPointer callData)
{
  int v,temp;
  v = ((XmScaleCallbackStruct *)callData)->value;
  switch ((long)clientData) {
  case DIAM: diam = (double)v / 100.0;
    break;
  case UPDATE: updateStep = v;
    break;
  case PLOTSCALE: plotdata.xscale = (double)v / 10.0;
    break;
  case ZOOMSCALE: 
    zoom.y = (double)v;
    zoom.x=zoom.y*wDrawSize.x/wDrawSize.y;
    
    temp = (int)((simArea.x - zoom.x) / 0.2) + 5;
    if (temp < 0) temp = 0;
    if (center.x > temp / 10) {
      XtVaSetValues(wHScrollB, XmNvalue, temp, NULL);
      center.x = (double)temp / 10.0;
    } 
    else if (center.x < -temp / 10) {
      XtVaSetValues(wHScrollB, XmNvalue, -temp, NULL);
      center.x = -(double)temp / 10.0;
    }
    XtVaSetValues(wHScrollB,
		  XmNminimum,     -temp,
		  XmNmaximum,     temp + 100,
		  NULL);
        
    temp = (int)((simArea.y - zoom.y) / 0.2) + 5;
    if (temp < 0)
      temp = 0;
    if (center.y > temp / 10) {
      XtVaSetValues(wVScrollB, XmNvalue, temp, NULL);
      center.y = (double)temp / 10.0;
    } 
    else if (center.y < -temp / 10) {
      XtVaSetValues(wVScrollB, XmNvalue, -temp, NULL);
      center.y = -(double)temp / 10.0;
    }
    XtVaSetValues(wVScrollB,
		  XmNminimum,     -temp,
		  XmNmaximum,     temp + 100,
		  NULL);
    
    break;
  }
  Redraw();
}


/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * scroll bar's value-changed callback (draw area scrollbars)
 */
void CbScroll(Widget w, XtPointer clientData, XtPointer callData)
{
  if((int)(long)clientData == HORIZONTAL)
    center.x = (double)((XmScrollBarCallbackStruct *)callData)->value / 10.0;
  else
    center.y = (double)((XmScrollBarCallbackStruct *)callData)->value / 10.0;

  Redraw();
}

/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

void UpdateScrollBars()
{
  int  temp = (int)((simArea.x*2 - zoom.x) / 0.2) + 5;
  if (temp < 0) temp = 0;
  if (center.x > temp / 10) {
    XtVaSetValues(wHScrollB, XmNvalue, temp, NULL);
    center.x = (double)temp / 10.0;
  } 
  else if (center.x < -temp / 10) {
    XtVaSetValues(wHScrollB, XmNvalue, -temp, NULL);
    center.x = -(double)temp / 10.0;
  }
  XtVaSetValues(wHScrollB,
		  XmNminimum,     -temp,
		  XmNvalue,(int)(center.x*10.0),
             	  XmNmaximum,     temp + 100,
		  NULL);
        
  temp = (int)((simArea.y - zoom.y) / 0.2) + 5;
  if (temp < 0)
    temp = 0;
  if (center.y > temp / 10) {
    XtVaSetValues(wVScrollB, XmNvalue, temp, NULL);
    center.y = (double)temp / 10.0;
  } 
  else if (center.y < -temp / 10) {
    XtVaSetValues(wVScrollB, XmNvalue, -temp, NULL);
    center.y = -(double)temp / 10.0;
  }
  XtVaSetValues(wVScrollB,
		  XmNminimum,     -temp,
                  XmNvalue,(int)(center.y*10.0),
		  XmNmaximum,     temp + 100,
		  NULL);
}

/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * drawingarea widget expose callback
 */
void CbExpose(Widget w, XtPointer clientData, XmDrawingAreaCallbackStruct *cbs)
{
  /* ensure last expose event in sequence */
  if (((XExposeEvent *)cbs->event)->count == 0)
    Refresh();
}


/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * plottingarea widget expose callback
 */
void CbExposePlot(Widget w, XtPointer clientData, 
		  XmDrawingAreaCallbackStruct *cbs)
{
  /* ensure last expose event in sequence */
  if (((XExposeEvent *)cbs->event)->count == 0)
    RefreshPlot(); /*also refreshes the range area*/
}

/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * plottingarea widget expose callback
 */
void CbExposenebPlot(Widget w, XtPointer clientData, 
		  XmDrawingAreaCallbackStruct *cbs)
{
  /* ensure last expose event in sequence */
  if (((XExposeEvent *)cbs->event)->count == 0)
    RefreshnebPlot(); 
}

/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/
/* Moves and resizes modify area selection if the dialog is open and calls 
   another function if the event is 3rd mouse button click.*/

void CbSelectBox(Widget wid, XEvent *event, String *args, int *num_args)
{
  double x,y,sx,sy,angle;
  int xpix,ypix;
  static VecD startp;
  char buff[MAXLINE];
  XButtonEvent *bevent = (XButtonEvent *) event;
  x = bevent->x;
  y = bevent->y;
  sx=wDrawSize.x/zoom.x;
  sy=wDrawSize.y/zoom.y;
  xpix=bevent->x;
  ypix=bevent->y;
  x=(double)( xpix-wDrawSize.x/2.0)/sx+center.x;
  y=(double)(-ypix+wDrawSize.y/2.0)/sy+center.y;
  

  if (*num_args != 1)
    XtError("CbSelectBox error!\n");

  /* start dragging the mouse with left mouse button pressed*/
  if ((strcmp(args[0],"down")==0) && wWinModFlag) {
    modarea.center.x=x;
      modarea.center.y=y;
      modarea.size.x=0.0;
      modarea.size.y=0.0;
      modarea.angle=0.0;
      startp.x=x;
      startp.y=y;
      Redraw();
  }
  /* stop dragging, area selection complete*/
  else if ((strcmp(args[0],"up")==0) && wWinModFlag) {
    
    sprintf(buff, "%g", modarea.center.x);
    XmTextSetString(c[MODAREA_X].text, buff);
    sprintf(buff, "%g", modarea.center.y);
    XmTextSetString(c[MODAREA_Y].text, buff);
    sprintf(buff, "%g", modarea.size.x);
    XmTextSetString(c[MODAREA_SX].text, buff);
    sprintf(buff, "%g", modarea.size.y);
    XmTextSetString(c[MODAREA_SY].text, buff);
    sprintf(buff, "%g", modarea.angle);
    XmTextSetString(c[MODAREA_ROT].text, buff);
    Redraw();
  }
  /* define selection by dragging the mouse */
  else if ((strcmp(args[0],"motion")==0) && wWinModFlag) {  
    if (x >= startp.x) {
      modarea.center.x = x-(x-startp.x)/2;
      modarea.size.x   = (x-startp.x);
    }
    else {
      modarea.center.x = x+(startp.x-x)/2;
      modarea.size.x   = (startp.x-x);
    }
    if (y <= startp.y) {
      modarea.center.y = y-(y-startp.y)/2;
      modarea.size.y   = -(y-startp.y);
    }
    else {
      modarea.center.y = y+(startp.y-y)/2;
      modarea.size.y   = -(startp.y-y);
    }
    Redraw();
  }
  /* move selection */
  else if ((strcmp(args[0],"smove")==0) && wWinModFlag) {
    /*modarea.center.x = x;
    modarea.center.y = y;
    Redraw();*/
    startp.x=x;
    startp.y=y;
  }
  else if ((strcmp(args[0],"move")==0) && wWinModFlag) {
    if (modrelatFlag) {
      modarea.center.x += (x-startp.x);
      modarea.center.y += (y-startp.y);
      startp.x=x;
      startp.y=y;
    } else {
      modarea.center.x = x;
      modarea.center.y = y;
    }
    Redraw();
  }
  else if ((strcmp(args[0],"endmove")==0) && wWinModFlag) {
    if (!modrelatFlag) {
      modarea.center.x = x;
      modarea.center.y = y;
    }
    sprintf(buff, "%g", modarea.center.x);
    XmTextSetString(c[MODAREA_X].text, buff);
    sprintf(buff, "%g", modarea.center.y);
    XmTextSetString(c[MODAREA_Y].text, buff);
    Redraw();
  }
  /* resize selection*/
  else if ((strcmp(args[0],"sresize")==0) && wWinModFlag) {
    if (modrelatFlag) {
      startp.x=x;
      startp.y=y;
    } else {
      modarea.size.x = fabs(x-modarea.center.x)*2.0;
      modarea.size.y = fabs(y-modarea.center.y)*2.0;
      Redraw();
    }
  }
  else if ((strcmp(args[0],"resize")==0) && wWinModFlag) {
    /*modarea.size.x = fabs(x-modarea.center.x)*2;
    modarea.size.y = fabs(y-modarea.center.y)*2;*/
    /*dist.x=x-modarea.center.x;
    dist.y=y-modarea.center.y;
    if (fabs(dist.x)>SMALLDOUBLE) { 
      modarea.size.x = VLen(dist)*cos(atan(dist.y/dist.x));
      modarea.size.y = VLen(dist)*sin(atan(dist.y/dist.x));
    }
    else {
    }*/
    if (modrelatFlag) {
      modarea.size.x = modarea.size.x + (x-startp.x);
      modarea.size.y = modarea.size.y + (y-startp.y);
      
      startp.x=x;
      startp.y=y;
      if (modarea.size.x < 0.0) {
	modarea.size.x = 0;
      }
      if (modarea.size.y < 0.0) {
	modarea.size.y = 0;
      }
    } else {
      modarea.size.x = fabs(x-modarea.center.x)*2.0;
      modarea.size.y = fabs(y-modarea.center.y)*2.0;


    }
    Redraw();
  }
  else if ((strcmp(args[0],"endresize")==0) && wWinModFlag) {
    /*modarea.size.x = fabs(x-modarea.center.x)*2;
    modarea.size.y = fabs(y-modarea.center.y)*2;*/
    sprintf(buff, "%g", modarea.size.x);
    XmTextSetString(c[MODAREA_SX].text, buff);
    sprintf(buff, "%g", modarea.size.y);
    XmTextSetString(c[MODAREA_SY].text, buff);
    Redraw();
  }
  /* rotate selection */
  else if ((strcmp(args[0],"srotate")==0) && wWinModFlag) {
    if (modrelatFlag) { 
      startp.x=x;
      startp.y=y;
    } else {
      if (fabs(x-modarea.center.x) < SMALLDOUBLE)
	if (y > modarea.center.y)
	  modarea.angle=0.0;
	else
	  modarea.angle=180.0;
      else {
	angle=atan((y-modarea.center.y)/(x-modarea.center.x));
	if (x < modarea.center.x)
	  angle += Pi;
	modarea.angle = angle * 180/Pi - 90.0;
      }
      Redraw(); 
    }
  }

  else if ((strcmp(args[0],"rotate")==0) && wWinModFlag) {
    if (modrelatFlag) {    
      modarea.angle = modarea.angle - (x-startp.x)*10;
    
      startp.x=x;
      startp.y=y;
      
      if (modarea.angle < -360.0) {
	modarea.angle = 0;
      }
      if (modarea.angle > 360.0) {
	modarea.angle = 0;
      }
    } else {
      if (fabs(x-modarea.center.x) < SMALLDOUBLE)
	if (y > modarea.center.y)
	  modarea.angle=0.0;
	else
	  modarea.angle=180.0;
      else {
	angle=atan((y-modarea.center.y)/(x-modarea.center.x));
	if (x < modarea.center.x)
	  angle += Pi;
	modarea.angle = angle * 180/Pi - 90.0;
      }
    }
    Redraw(); 
  }
  else if ((strcmp(args[0],"endrotate")==0) && wWinModFlag) {
    
    sprintf(buff, "%g", modarea.angle);
    XmTextSetString(c[MODAREA_ROT].text, buff);
    Redraw(); 
  }
  /* scroll using ctrl+right mouse button*/ /*partially implemented*/
  else if ((strcmp(args[0],"sscroll")==0)) {
    startp.x=x;
    startp.y=y;
  }
  else if ((strcmp(args[0],"scroll")==0)) {
    
    center.x = center.x - (x-startp.x)*0.65;
    center.y = center.y - (y-startp.y)*0.65;
    
    startp.x=x;
    startp.y=y;
    UpdateScrollBars();
    Redraw();
  }
  /* zoom using ctrl+right mouse button*/
  else if ((strcmp(args[0],"szoom")==0)) {
    startp.x=x;
    startp.y=y;
  }
  else if ((strcmp(args[0],"zoom")==0)) {
    
    zoom.y = zoom.y + (y-startp.y)*1.3;
    zoom.x=zoom.y*wDrawSize.x/wDrawSize.y;

    if (zoom.y < 1.0) zoom.y =1.0; 
    if (zoom.y > 200.0) zoom.y = 200.0;

    startp.x=x;
    startp.y=y;
    
    Redraw();
  }
  else if ((strcmp(args[0],"endzoom")==0)) {
    XtVaSetValues(wZoomScale,XmNvalue,(int)zoom.y,NULL);
    UpdateScrollBars();
  }
  /* print atom info */
  else if (strcmp(args[0],"info")==0)  
    /*do not alter selection, print atom info (mouse button 3)*/
    CbPrintAtomInfo(bevent->x,bevent->y);
}




/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

/*
 * Drawing area mouse click callback:
 * print the nearest atom information.
 */
void CbPrintAtomInfo(double xpix, double ypix/*Widget w, XtPointer clientData, 
		  XmDrawingAreaCallbackStruct *cbs*/)
{
  int /*xpix,ypix,*/j,c2,k,offset;
  /*int c;*not in use*/
  double x,y,sx,sy,dr,drmin,dx,dy;
  static char *aname[5]={"Bf","Af","??","A","B"};
  VecD rs,invWid,xv,iof[] = {{0,0},{1,0},{1,1},{0,1},{-1,1}};
  VecI cc,cc2;

  if (/*cbs->event->xany.type==ButtonPress*/1) {

    sx=wDrawSize.x/zoom.x;
    sy=wDrawSize.y/zoom.y;
    /*xpix=cbs->event->xbutton.x;
    ypix=cbs->event->xbutton.y;*/
    x=( xpix-wDrawSize.x/2.0)/sx+center.x;
    y=(-ypix+wDrawSize.y/2.0)/sy+center.y;
    xv.x=x;
    xv.y=y;

    /* Get the linear index of the current cell to c */
    VDiv(invWid, cells, simArea);
    VSAdd(rs,xv,0.5,simArea);
    VMul(cc,rs,invWid);
    /*c=VLinear(cc,cells)+nAtomOrig;*not in use*/
    drmin=1.0e20;
    k=-1;
    /* Go through atoms in the current and neighboring cells */
    for (offset = 0; offset < 5; offset ++) {
      VAdd(cc2,cc,iof[offset]);
      if (cc2.x>=0 && cc2.x<cells.x && cc2.y>=0 && cc2.y<cells.y) {
	c2=VLinear(cc2,cells)+nAtomOrig;
	for (j=cellList[c2];j>=0;j=cellList[j]) {
	  dx=ae[j].r.x-xv.x;
	  dy=ae[j].r.y-xv.y;
	  dr=(dx*dx+dy*dy);
	  if (dr<drmin) {
	    k=j;
	    drmin=dr;
	  }
	}
      }
    }
    fprintf(stderr,"Pos %g %g ",x,y);
    if (k>=0) {
      fprintf(stderr," atom %d (%s) at %g %g v %g %g Epot %g",
	      k,aname[ae[k].atype+2], ae[k].r.x,ae[k].r.y, 
	      ae[k].rv.x, ae[k].rv.y,
	      ae[k].potEn);
      if (sinfo==SXX || sinfo==SXY || sinfo==SYY) 
	fprintf(stderr," stress %g",ae[k].stressij);
    }
    fprintf(stderr,"\n");
  }
}

/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/

/*needed by the notorious tracking dialog*/
void CbSaveTra(Widget wTextSave,XtPointer client_data,XtPointer call_data)
{
  int i;
  char *tulostus = XmTextFieldGetString(wTextSave);

  for(i=0; i < 5; i++){
    numero[i] = -1;
  }
 
  sscanf(tulostus,"%d,%d,%d,%d,%d",&numero[0],&numero[1],
	 &numero[2],&numero[3],&numero[4]); 
  printf(" %d,%d,%d,%d,%d \n",numero[0],numero[1],numero[2],
	 numero[3],numero[4]);
  
}

/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/
/*this function is called everytime a dialog is closed either by the window
  manager or by the application*/
void CbCloseDialog(Widget widget, XtPointer client_data, XtPointer call_data) 
{
  Widget reset_widget = (Widget) client_data;
  XtSetSensitive(reset_widget, True);
  if (widget != NULL) {
    if ((wSeamMan != NULL) && (widget == XtParent(wSeamMan)))
      wWinSeamFlag = False;
    else if ((wParamMan != NULL) && (widget == XtParent(wParamMan)))  {
      if(wWinSeamFlag) XtPopdown(XtParent(wSeamMan));
      strainFlag = False;
      CalcAtoms();
      InitRun();
      XtSetSensitive(wRunPushB, True);
      Redraw();
      //RedrawPlot(1);
    }
    else if ((wTrackMan != NULL) && (widget == XtParent(wTrackMan)))
      wWinTrackFlag=0;
    else if ((wAniMan != NULL) && (widget == XtParent(wAniMan)))
      wWinAniFlag=0;
    else if ((wnebMan != NULL) && (widget == XtParent(wnebMan)))
      wWinnebFlag=0;
    else if ((wModMan != NULL) && (widget == XtParent(wModMan))) {
      wWinModFlag= False;
      Redraw();
    }
    else if ((wEventsMan != NULL) && (widget == XtParent(wEventsMan)))
      wWinEventsFlag=0;
  }
}

/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/

/*callback function for all menus in menubar or display menu*/
void CbMenu(Widget menu_item,XtPointer client_data,XtPointer call_data)
{
  switch ((int) client_data) {
  case M_NEW:  /*new simulation*/
    wWinSeamFlag = 0;
    initFlag = 0;
    if (runningFlag) { /* stop a running simulation */
      XtRemoveWorkProc(id);
      runningFlag = False;
      XtVaSetValues(wRunPushB, XmNalignment,XmALIGNMENT_CENTER, 
		    XtVaTypedArg, XmNlabelString, T_STRING("Run"),
		    NULL);
    }
    if (sinfo == SVECT) {
      sinfo = STYPE;
    }
    if (savetscheme)
      fclose(tsavefile);
    cool.programon = 0;
    ani.snaphelp  = 0;
    XtSetSensitive(wRunPushB, False);
    SaveParams(STORE);
    CalcAtoms();
    OpenDialog(wParamMan,menu_item,"param");
    XtSetSensitive(wPushBCancel, True);
    Redraw();
    //RedrawPlot(0);
    break;
  case M_SNAPSHOT:   /*save snapshot*/
    SnapShot();
    break;
  case M_ANIMATION:   /*save animation*/
    wWinAniFlag = True;
    if (runningFlag) { /* stop a running simulation*/
      XtRemoveWorkProc(id);
      runningFlag = False;
      /*XtSetSensitive(wStrainPushB, False);*/
      XtVaSetValues(wRunPushB,  XmNalignment,XmALIGNMENT_CENTER,
		    XtVaTypedArg,XmNlabelString, T_STRING("Run"),
		    NULL);
    }
    /*if (vectorFlag) {
      vectorFlag = False;*/
    /*XtVaSetValues(wVectorToggle, XmNset, False, NULL);
      }*/
    OpenDialog(wAniMan, menu_item,"ani");
    wWinAniFlag = True;
    Redraw();
    break;
  case M_EXIT:  /*bexit*/
    bexit(0);
    break;

  case M_STRAIN: /*strain and pressure*/ 
    OpenDialog(wStrainMan, menu_item,"strain");
    /*Redraw();*/
    break;
  case M_TEMPERATURE: /*temperature*/
    OpenDialog(wTempMan, menu_item,"temp");
    /*Redraw();*/
    break;
  case M_NEWCRACK: /*make new crack*/
    /*CreateInitialCrack(r,&nAtom,crack.center.x,crack.center.y,
		       crack.size.x,crack.size.y,crack.angle);*/
    /*modinfo=M_MODRM;*/
    /*OpenModDialog(menu_item,client_data,call_data);*/
    OpenDialog(wModMan,menu_item,"mod");
    wWinModFlag = True;
    Redraw();
    break;
  case M_RESCALE:
    plotdata.rescale=1;
    break;
  case M_ZOOM: /*zooming*/ 
    OpenDialog(wZoomMan, menu_item,"zoom");
    break;
  case M_TEST: /*testing*/ 
    OpenDialog(wTestMan, menu_item,"test");
    break;
  case M_COLOR: /*color ranges*/ 
    OpenDialog(wColorMan, menu_item,"color");
    break;

  case M_TRACK:
    if (sinfo == SVECT) {
      sinfo = SKIN;
      /*XtVaSetValues(wVectorToggle, XmNset, False, NULL);*/
    }
    OpenDialog(wTrackMan, menu_item,"track");
    wWinTrackFlag = True;
    /*Redraw();*/
    break;
  case M_NEB:
    wWinnebFlag = True;
    if (images > 1) {
      int i,j; 
      for (i=1;i<images;i++) {
	for (j=0;j<nAtom;j++) {
	  ae[nAtom*i+j]=ae[j];
	}
      }
      if (readatoms) {
	for (i=1; i < images; i++) {
	  LoadImageFromFile(i);
	}
      }
    }
    OpenDialog(wnebMan, menu_item,"neb");
    break;

  case M_EVENTS:
    wWinEventsFlag = True;
    OpenDialog(wEventsMan, menu_item,"events");
    break;
    
  case M_RESETTIME:
    stepCount=0; 
    break;


    /*below this are display -option menu items*/

  case M_KIN:
    TASensitive(AVRCUT, False);
    sinfo = SKIN;
    Redraw();
    break;
  case M_POT:
    TASensitive(AVRCUT, False);
    sinfo = SPOT;
    Redraw();
    break;
  case M_SIGMAXX:
    TASensitive(AVRCUT, False);
    sinfo = SXX;
    Redraw();
    break;
  case M_SIGMAXY:
    TASensitive(AVRCUT, False);
    sinfo = SXY;
    Redraw();
    break;
  case M_SIGMAYY:
    TASensitive(AVRCUT, False);
    sinfo = SYY;
    Redraw();
    break;
  case M_ATOMTYPE:
    TASensitive(AVRCUT, False);
    sinfo = STYPE;
    Redraw();
    break;
  case M_FORCE:
    sinfo = SFORCE;
    TASensitive(AVRCUT, False);
    Redraw();
    break;  
  case M_VECTORS:
    sinfo = SVECT;
    TASensitive(AVRCUT, True);
    Redraw();
    break; 
  case M_MODRM:
    modinfo=M_MODRM;
    TASensitive(MODVX, False);
    TASensitive(MODVY, False);
    break;
  case M_MODSTOP:
    modinfo=M_MODSTOP;
    TASensitive(MODVX, False);
    TASensitive(MODVY, False);
    break;
  case M_MODTYPE1:
    modinfo=M_MODTYPE1;
    TASensitive(MODVX, False);
    TASensitive(MODVY, False);
    break;
  case M_MODTYPE2:
    modinfo=M_MODTYPE2;
    TASensitive(MODVX, False);
    TASensitive(MODVY, False);
    break;
  case M_MODVEL:
    modinfo=M_MODVEL;
    TASensitive(MODVX, True);
    TASensitive(MODVY, True);
    break;  
  case M_MODFIX:
    modinfo=M_MODFIX;
    TASensitive(MODVX, True);
    TASensitive(MODVY, True);
    break;  
  case M_MODFIXX:
    modinfo=M_MODFIXX;
    TASensitive(MODVX, True);
    TASensitive(MODVY, False);
    break;  
  case M_MODFIXY:
    modinfo=M_MODFIXY;
    TASensitive(MODVX, False);
    TASensitive(MODVY, True);
    break;  
  case M_MODRMF:
    modinfo=M_MODRMF;
    TASensitive(MODVX, False);
    TASensitive(MODVY, False);
    break;  
  case M_MODFORCE:
    modinfo=M_MODFORCE;
    TASensitive(MODVX, True);
    TASensitive(MODVY, True);
    break;
  case M_MODMOVE:
    modinfo=M_MODMOVE;
    TASensitive(MODVX, True);
    TASensitive(MODVY, True);
    break;
 case M_MODTAG:
    modinfo=M_MODTAG;
    TASensitive(MODVX, True);
    TASensitive(MODVY, False);
    break;
 case M_MODROT:
    modinfo=M_MODROT;
    TASensitive(MODVX, True);
    TASensitive(MODVY, False);
    break;
  }

}

/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/

void CbList(Widget menu_item,XtPointer client_data,XtPointer call_data)
{
  XmListCallbackStruct *cbs = (XmListCallbackStruct *) call_data;
  char *choice;
  
  XmStringGetLtoR(cbs->item, XmFONTLIST_DEFAULT_TAG,&choice);
  /*fprintf(stderr,"item: %s\n",choice);*/
  strcpy(tfname,choice);
}

/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/

void UpdateTestText(char *string)
{
  XmTextSetString(wTextTest, string);
}

void dtest(double *var, char *varname, char *string, int gettest, 
	   double value) {
  char text[MAXLINE];
  if (!strcmp(varname,string)) {
    if (gettest) {
      sprintf(text,"%s=%f",varname,*var);
      UpdateTestText(text);
    } else {
      *var=value;
      sprintf(text,"%s=%f",varname,*var);
      UpdateTestText(text);
    }
  }
}

void itest(int *var, char *varname, char *string, int gettest, 
	   int value) {
  char text[MAXLINE];
  if (!strcmp(varname,string)) {
    if (gettest) {
      sprintf(text,"%s=%d",varname,*var);
      UpdateTestText(text);
    } else {
      *var=value;
      sprintf(text,"%s=%d",varname,*var);
      UpdateTestText(text);
    }
  }
}



void CbTextActivate(Widget wText, XtPointer client_data,
                                      XtPointer call_data)
{
  char *string = XmTextFieldGetString(wText);
  int length=strlen(string);
  char text[MAXLINE],varname[MAXLINE];
  int number,settest,gettest;
  double dvalue;
  int ivalue;
  XmTextFieldSetString(wText,"");
  settest=!strncmp(string,"set",3);
  gettest=!strncmp(string,"get",3);
  if ((length>4) && (settest || gettest)) {
    
    sscanf(string,"%s %s %lf",text, varname,&dvalue);
    sscanf(string,"%s %s %d",text, varname,&ivalue);
    dtest(&nebrepA,"nebrepA",varname,gettest,dvalue);
    dtest(&nebrepA,"nebrepAlpha",varname,gettest,dvalue);
    itest(&nebrepFlag,"nebrepFlag",varname,gettest,ivalue);
  } else if (!strcmp(string,"command")) {
    UpdateTestText("command");
  }
}
