#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>  /*DBL_MAX*/ 
#include <float.h>   /*DBL_MAX*/

/* time */
#include <sys/types.h>
#include <time.h>
#include <sys/timeb.h>

#include "control.h"
#include "graphics.h"
#include "simu.h"
#include "fitsigma.h"
#include "calc.h"
#include "callback.h"
#include "gvar.h"    /*load global variables*/
#include "track.h"
#include "r250.h"
#include "conjgrad.h"

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void cgStep(VecD *dir, double step) 
{
  int i;
  double Ek;
  kinEnmin = 1.0e20;
  kinEnmax = -1.0e20;

  for (i=cImage*nAtom; i<(cImage+1)*nAtom; i++) {
    Ek = 0.0;
    if (!(ae[i].fixvel & FIXX)) {
      ae[i].r.x += dir[i-cImage*nAtom].x*step;
      ae[i].rv.x = dir[i-cImage*nAtom].x*step;
      Ek += SQ(dir[i-cImage*nAtom].x*step);
    }
    if (!(ae[i].fixvel & FIXY)) {
      ae[i].r.y += dir[i-cImage*nAtom].y*step;
      ae[i].rv.y = dir[i-cImage*nAtom].y*step;
      Ek += SQ(dir[i-cImage*nAtom].y*step);
    }
    Ek = sqrt(Ek);
    //Ek *= 0.5;
    if (Ek < kinEnmin) kinEnmin = Ek;
    if (Ek > kinEnmax) kinEnmax = Ek;
  }
  ComputeForces();
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

double cgPot(VecD *dir, double step) 
{
  int i;
  double pot1, pot0=potEnergy;
  static VecD *x0=NULL;
  if (x0 == NULL)
    ALLOC(x0,nAtom,VecD);
  
  for (i=cImage*nAtom; i<(cImage+1)*nAtom; i++) {
    x0[i-cImage*nAtom]=ae[i].r;
    if (!(ae[i].fixvel & FIXX)) ae[i].r.x += dir[i-cImage*nAtom].x*step;
    if (!(ae[i].fixvel & FIXY)) ae[i].r.y += dir[i-cImage*nAtom].y*step;
  
  }
  ComputeForces();
  pot1=potEnergy;
  for (i=cImage*nAtom; i<(cImage+1)*nAtom; i++) {
    ae[i].r = x0[i-cImage*nAtom];
  }
  
  potEnergy=pot0;
  return pot1;
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

double cgDir(VecD *dir, int init)
{
  double nom=0.0,denom=0.0;
  int i;
  double length=0.0;
  static int k=0;
  static VecD *prevdir=NULL;
  if (prevdir == NULL) 
    ALLOC(prevdir,nAtom,VecD);
  
  if ((k>0) && !init) {
    for (i=0; i<nAtom; i++) {
      nom += (ae[i+cImage*nAtom].ra.x-cgprevgrad[i].x)*ae[i+cImage*nAtom].ra.x;
      nom += (ae[i+cImage*nAtom].ra.y-cgprevgrad[i].y)*ae[i+cImage*nAtom].ra.y;
      denom += cgprevgrad[i].x*cgprevgrad[i].x;
      denom += cgprevgrad[i].y*cgprevgrad[i].y;
    }
    if (fabs(denom) > SMALLDOUBLE) {
      for (i=0; i<nAtom; i++) {
	dir[i].x = ae[i+cImage*nAtom].ra.x + nom/denom*prevdir[i].x;
	dir[i].y = ae[i+cImage*nAtom].ra.y + nom/denom*prevdir[i].y;
	prevdir[i] = dir[i];
	cgprevgrad[i] = ae[i+cImage*nAtom].ra;
	length += SQ(dir[i].x);
	length += SQ(dir[i].y);
      }
    } else {
      length = 0.0;
    }
  } else {
    //fprintf(stderr," k=0 ");
    k=0;
    ComputeForces();
    potEnergy /= nAtom;
    for (i=0; i<nAtom; i++) {
      dir[i] = ae[i+cImage*nAtom].ra;
      prevdir[i] = dir[i];
      cgprevgrad[i] = dir[i];
      length += SQ(dir[i].x);
      length += SQ(dir[i].y);
    }
  }
  k++;
  length=sqrt(length);
  return length;
}


/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

#define NRANSI
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-30 //1.0e-20
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);
static double maxarg1,maxarg2;
#define FMAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
        (maxarg1) : (maxarg2))

void cgmnbrak(double *ax, double *bx, double *cx, double *fa, double *fb, 
	    double *fc, VecD *dir)
{
  double ulim,u,r,q,fu,dum;
  
  *fa=cgPot(dir,*ax);
  *fb=cgPot(dir,*bx);
  if (*fb > *fa) {
    SHFT(dum,*ax,*bx,dum)
      SHFT(dum,*fb,*fa,dum)
      }
  *cx=(*bx)+GOLD*(*bx-*ax);
  *fc=cgPot(dir,*cx);
  while (*fb > *fc) {
    r=(*bx-*ax)*(*fb-*fc);
    q=(*bx-*cx)*(*fb-*fa);
    u=(*bx)-((*bx-*cx)*q-(*bx-*ax)*r)/
      (2.0*SIGN(FMAX(fabs(q-r),TINY),q-r));
    ulim=(*bx)+GLIMIT*(*cx-*bx);
    if ((*bx-u)*(u-*cx) > 0.0) {
      fu=cgPot(dir,u);
      if (fu < *fc) {
	*ax=(*bx);
	*bx=u;
	*fa=(*fb);
	*fb=fu;
	return;
      } else if (fu > *fb) {
	*cx=u;
	*fc=fu;
	return;
      }
      u=(*cx)+GOLD*(*cx-*bx);
      fu=cgPot(dir,u);
    } else if ((*cx-u)*(u-ulim) > 0.0) {
      fu=cgPot(dir,u);
      if (fu < *fc) {
	SHFT(*bx,*cx,u,*cx+GOLD*(*cx-*bx))
	  SHFT(*fb,*fc,fu,cgPot(dir,u))
	  }
    } else if ((u-ulim)*(ulim-*cx) >= 0.0) {
      u=ulim;
      fu=cgPot(dir,u);
    } else {
      u=(*cx)+GOLD*(*cx-*bx);
      fu=cgPot(dir,u);
    }
    SHFT(*ax,*bx,*cx,u)
      SHFT(*fa,*fb,*fc,fu)
      }
}
#undef GOLD
#undef GLIMIT
#undef TINY
#undef SHFT
#undef NRANSI


#define NRANSI
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define ITMAX 100
#define CGOLD 0.3819660
//#define ZEPS 1.0e-10
#define ZEPS 1.0e-30
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);


/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

double cgbrent(double ax, double bx, double cx, double tol,
	    double *xmin, VecD *dir)
{
  int iter;
  double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
  double e=0.0;
  
  a=(ax < cx ? ax : cx);
  b=(ax > cx ? ax : cx);
  x=w=v=bx;
  fw=fv=fx=cgPot(dir,x);
  for (iter=1;iter<=ITMAX;iter++) {
    xm=0.5*(a+b);
    tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
    if ((fabs(x-xm) <= (tol2-0.5*(b-a))) || (iter >= cgMaxLMSteps)) {
      *xmin=x;
      /* fprintf(stderr," br.it:%d  ",iter); */
      return fx;
    }
    if (fabs(e) > tol1) {
      r=(x-w)*(fx-fv);
      q=(x-v)*(fx-fw);
      p=(x-v)*q-(x-w)*r;
      q=2.0*(q-r);
      if (q > 0.0) p = -p;
      q=fabs(q);
      etemp=e;
      e=d;
      if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
	d=CGOLD*(e=(x >= xm ? a-x : b-x));
      else {
	d=p/q;
	u=x+d;
	if (u-a < tol2 || b-u < tol2)
	  d=SIGN(tol1,xm-x);
      }
    } else {
      d=CGOLD*(e=(x >= xm ? a-x : b-x));
    }
    u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
    fu=cgPot(dir,u);
    if (fu <= fx) {
      if (u >= x) a=x; else b=x;
      SHFT(v,w,x,u)
	SHFT(fv,fw,fx,fu)
	} else {
	  if (u < x) a=u; else b=u;
	  if (fu <= fw || w == x) {
	    v=w;
	    w=u;
	    fv=fw;
	    fw=fu;
	  } else if (fu <= fv || v == x || v == w) {
	    v=u;
	    fv=fu;
	  }
	}
  }
  //nrerror("Too many iterations in brent");
  fprintf(stderr,"brent-error->exiting...\n");
  bexit(-1);
  //return fx;
}
#undef SIGN
#undef ITMAX
#undef CGOLD
#undef ZEPS
#undef SHFT
#undef NRANSI


#define R 0.61803399
#define C (1.0-R)
#define SHFT2(a,b,c) (a)=(b);(b)=(c);
#define SHFT3(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);


/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void cggolden(double ax, double bx, double cx, double tol, VecD *dir)
{
  double f1,f2,x0,x1,x2,x3;
  int i=0;

  x0=ax;
  x3=cx;
  if (fabs(cx-bx) > fabs(bx-ax)) {
    x1=bx;
    x2=bx+C*(cx-bx);
  } else {
    x2=bx;
    x1=bx-C*(bx-ax);
  }
  /*f1=(*f)(x1);
  f2=(*f)(x2);*/
  f1=cgPot(dir,x1);
  f2=cgPot(dir,x2);
  while ((fabs(x3-x0) > tol*(fabs(x1)+fabs(x2))) && (i<25)) {
    i++;
    if (f2 < f1) {
      SHFT3(x0,x1,x2,R*x1+C*x3)
	SHFT2(f1,f2,cgPot(dir,x2))
	} else {
	  SHFT3(x3,x2,x1,R*x2+C*x0)
	    SHFT2(f2,f1,cgPot(dir,x1))
	    }
  }
  /* fprintf(stderr,"golden i:%d,   ",i); */
  if (f1 < f2) {
    cgStep(dir, x1); 
    /**xmin=x1;
    return f1;*/
  } else {
    cgStep(dir, x2); 
    /**xmin=x2;
    return f2;*/
  }
}
#undef C
#undef R
#undef SHFT2
#undef SHFT3

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void CompareOldState()
{
  int i;
  static int counter=0;
  static VecD *xr=NULL;
  double diff=0.0;
  if (xr == NULL) {
    ALLOC(xr,nAtom,VecD);
    for (i=cImage*nAtom; i<(cImage+1)*nAtom; i++) {
      xr[i-cImage*nAtom]=ae[i].r;
    }
  }

  if (counter > 1) {
    counter=0;
    for (i=0; i<nAtom; i++) {
      diff += SQ(xr[i].x - ae[i+cImage*nAtom].r.x);
      diff += SQ(xr[i].y - ae[i+cImage*nAtom].r.y);
    }
    diff = sqrt(diff);
    /* fprintf(stderr,"\ncomparetest:%f\n",diff); */
    for (i=cImage*nAtom; i<(cImage+1)*nAtom; i++) {
      xr[i-cImage*nAtom]=ae[i].r;
    }
  } 
  counter++;
  
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/



void cgMinimizer()
{
  double ax,bx,cx,fa,fb,fc;
  double xmin;
  static double step=0.002; /*prev:0.001*/
  double defstep=0.002;
  double realstep;
  int i,iter;
  double pot0,pot1;
  double length;
  
  static int init=0;
  static VecD *xr=NULL;
  static VecD *xa=NULL;
  static VecD *dir=NULL;
  //static int lastImage=0;
  double cgPEeps=0.000001;
  static int convIter=0;
  static int semiLMIter=0;
  static int semiMinNow=0;
  double currentStep=0.0;

  
  if (xr == NULL)
    ALLOC(xr,nAtom,VecD);
  if (xa == NULL)
    ALLOC(xa,nAtom,VecD);
  if (dir == NULL) 
    ALLOC(dir,nAtom,VecD);


  //CompareOldState();
  
  /***********************************/
  /* steepest descent */
  if (semisteepFlag) {
    length=0.0;
    for (i=cImage*nAtom; i<(cImage+1)*nAtom; i++) {
      dir[i-cImage*nAtom]=ae[i].ra;
      length += SQ(ae[i].ra.x);
      length += SQ(ae[i].ra.y);
    }
    length=sqrt(length);
    /***********************************/
  } else {
    /* conjugate gradient */
    //if (lastImage != cImage) init=1;
    if (cgInit) {
      /* fprintf(stderr,"cgInit=1\n"); */
      cgInit=0;
      init=1;
    }

    length=cgDir(dir,init);
    init=0;
    //lastImage=cImage;
  }    
  pot0=potEnergy*nAtom;
  
  //fprintf(stderr,"length: %f  ",length);
  if (length < SMALLDOUBLE) {
    stopFlag=1;
    cgInit=1;
    fprintf(stderr,"too short dir vector in cg ");
    ComputeForces();
    return;
  }
  
  /* semi line minim. */
  if (cgSemiLMFlag) {
    if (semiLMIter > 10) {
      semiLMIter=0;
      semiMinNow=1;
      /* fprintf(stderr,"semiLM now\n"); */
    } else
      semiLMIter++;
  }

  if (cgLineMinFlag || semiMinNow) { /* cg or sd with line minimization */
    if (semiMinNow) {
      semiMinNow=0;
      semiLMIter=0;
    }
    ax=0.0;
    bx=0.00013;
    cx=0.002618;
    if (cgBracketFlag) {
      cgmnbrak(&ax,&bx,&cx,&fa,&fb,&fc, dir);
      if (!((fb<fa) && (fb<fc))) {
	stopFlag=1;
	fprintf(stderr,"cg: unable to bracket a minimum!\n");
      }
    }
    //fprintf(stderr,"Bracket: ax:%f, bx:%f, cx: %f\n",ax,bx,cx);
    //cggolden(ax, bx, cx, 0.00001,dir);
    cgbrent(ax, bx, cx, 0.00001,&xmin,dir);
    cgStep(dir,xmin);
    step = xmin; 
    currentStep = xmin * length;
/*     fprintf(stderr,"Bracket: %f, xmin: %f, dir: %f, step: %f\n", */
/* 	    cx-bx,xmin,length,length*xmin); */
  } else { /* cg or sd without line minimization */
    realstep=step /*/ length*/;
    
    for (i=cImage*nAtom; i<(cImage+1)*nAtom; i++) {
      xr[i-cImage*nAtom] = ae[i].r;
      xa[i-cImage*nAtom] = dir[i-cImage*nAtom];
      
      
      //if (!(ae[i].fixvel & FIXX)) ae[i].r.x += xa[i-cImage*nAtom].x*realstep;
      //if (!(ae[i].fixvel & FIXY)) ae[i].r.y += xa[i-cImage*nAtom].y*realstep;
    }
    cgStep(xa, realstep);
    ComputeForces();
    
    if (potEnergy > pot0) {
      iter=0;
      while (potEnergy > pot0) {
		
	if (iter>20) {
	  for (i=cImage*nAtom; i<(cImage+1)*nAtom; i++) {
	    ae[i].r = xr[i-cImage*nAtom];
	  }
	  ComputeForces();
	  fprintf(stderr,"too many iterations in cg/sd without line min. ");
	  step=defstep;
	  currentStep = 0.0;
	  kinEnergy = currentStep;
	  if (cgSemiLMFlag)
	    semiMinNow=1;
	  cgInit=1;
	  //stopFlag=1;
	  return;
	}
	
	step *= 0.5;
	realstep=step/*/length*/;
	currentStep = realstep * length;

	for (i=cImage*nAtom; i<(cImage+1)*nAtom; i++) {  
	  ae[i].r = xr[i-cImage*nAtom];
	  /*if (!(ae[i].fixvel & FIXX)) 
	    ae[i].r.x += xa[i-cImage*nAtom].x*realstep;
	  if (!(ae[i].fixvel & FIXY)) 
	    ae[i].r.y += xa[i-cImage*nAtom].y*realstep;*/
	}
	cgStep(xa,realstep);
	ComputeForces();
	iter++;
      } 
      /* fprintf(stderr,"cg without linemin, iter: %d, step:%f\n",iter,step); */
    } else {
      currentStep = realstep * length;
      kinEnergy = currentStep;
      if (!cgSemiLMFlag)
	step *= 1.03;

    }
/*     fprintf(stderr,"cgNoLineMin dir: %f, step: %f, real step: %f\n",length, */
/* 	    realstep, currentStep); */
  }
  
  //fprintf(stderr,"step=%f\n",step);
  
  if (fabs(potEnergy-pot0) < cgPEeps) {
    if (convIter > 50) {
      fprintf(stderr, "cg converged: Delta(Epot)=%f\n",potEnergy-pot0);
      stopFlag=1;
      convIter=0;
    } else {
      convIter++;

    }
  } else
    convIter=0;

  if ((step < 0.000005) && !cgLineMinFlag) { 
    /* fprintf(stderr,"cg:very small step->reinit\n"); */
    step = defstep;
    init=1;
    if (cgSemiLMFlag) semiMinNow=1;
    //return;
  }
  kinEnergy = currentStep;
}


/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/



