
/* The following string is for emacs time-stamp feature            */
/* It is updated every time you save the file in emacs (or Xemacs) */
/* and have the time-stamp feature set.                            */
char *build="Time-stamp: <2003/06/30 10:09:38 akuronen>";
char *progname="boundary2d";
#define VERSION "NEB 1.1"

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <signal.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>        /*function toupper(...)*/

#include <Xm/DrawingA.h>

#include "graphics.h"
#include "control.h"
#include "callback.h"
#include "calc.h"
#include "fitsigma.h"
#include "simu.h"

#define MAINFILE
#include "gvar.h"  /*global variables*/
#undef MAINFILE

extern Container c[];

/*--------------------------------------------------------------*/
/* These are used in T scheme == .events */

#define VOIDTEMP -1.0
#define LARGETIME 1.0E20
#define MAXTSCHEME 500

#define TschemeUnknown    0
#define TschemeTscaleOn   1
#define TschemeTscaleOff  2
#define TschemeRandomize  3
#define TschemeQuenchOn   4
#define TschemeQuenchOff  5
#define TschemeQuit       6
#define TschemeModifyArea 7

#define STschemeTscaleOn   "TSCALEON"
#define STschemeTscaleOff  "TSCALEOFF"
#define STschemeRandomize  "RANDOMIZE"
#define STschemeQuenchOn   "QUENCHON"
#define STschemeQuenchOff  "QUENCHOFF"
#define STschemeQuit       "QUIT"
#define STschemeModifyArea "MODIFYAREA"

char *TschemeNames[]={"*Unknown*","TscaleOn ","TscaleOff","Randomize",
		      "QuenchOn ","QuenchOff","Quit     ","ModifyArea"};
int iTschemeMax = 0;	 
double tTscheme[MAXTSCHEME]; /* times of T scheme events */
int evTscheme[MAXTSCHEME];   /* T scheme events  */
double TTscheme[MAXTSCHEME]; /* temperatures of T scheme events and subtypes
				of modify area events */
RArea ModAT[MAXTSCHEME];     /* areas of T scheme modify events */
VecD ModXYT[MAXTSCHEME];     /* veloc./forces of T scheme mod. events*/ 



/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

/*
 * The Main Function of Boundary.
 */
int main(int argc, char **argv)
{
  char pfname[MAXLINE]="boundary."; /*TEST*/
  struct sigaction sact;
  int argc0;
  char **argv0;
  /*
    The following allows one to force the program
    to dump the configuration to disk by sending signal
    SIGHUP to the process. Just give the command
    
              kill -HUP <pid>.
  */
  
  sact.sa_sigaction=(*CatchHUP);
  sigemptyset(&sact.sa_mask);
  sact.sa_flags=SA_SIGINFO;
  sigaction(SIGHUP,  &sact, NULL); 
  sighupFlag=0;
    
  fprintf(stderr,"%s version %s %s\n",progname,VERSION,build+12);  

#if defined(HARMONIC) || defined(SHOWINDEX) || defined(DRAWHEXAGONS) || defined(ONEBODY)
  fprintf(stderr,"\nAdditional features:\n");
#endif
#ifdef HARMONIC
  fprintf(stderr,"  Interaction is harmonic (HARMONIC defined).\n");
#endif
#ifdef SHOWINDEX
  fprintf(stderr,"  Atomic index drawn (SHOWINDEX defined).\n");
#endif
#ifdef DRAWHEXAGONS
  fprintf(stderr,"  Atoms drawn as hexagons (DRAWHEXAGONS defined).\n");
#endif
#ifdef ONEBODY
  fprintf(stderr,"  One-body potential applied (ONEBODY defined).\n");
#endif
#if defined(HARMONIC) || defined(SHOWINDEX) || defined(DRAWHEXAGONS) || defined(ONEBODY)
  fprintf(stderr,"\n\n");
#endif

  SetDefaults();    /* Initial values for parameters */

/********************** TEST *********************/
  /*read data from file -> correct values for setupgraphics() so that the 
    widgets have correct initial states */
  
  XtSetup(argc,argv);
  argc=0;
  while (argv[argc] != NULL) {      /* recount command line options after*/
    argc++;                      /* Xt command line option parsing*/
  }

  getsomearguments(argc,argv);
  if (readatoms) {
   
    strcat(pfname,xyzloadnumber);
    strcat(pfname,".param");
    ROWParams("read",pfname);
    
  }
/************************ TEST ***********************/
  SetupGraphics();     /* Create the main window of the program,
				    initialize the Xt Toolkit and
				    get Xt command line options */
  
  argc=0;
  while (argv[argc] != NULL) {      /* recount command line options after*/
    argc++;                      /* Xt command line option parsing*/
  }
  
  //getsomearguments(argc,argv);   /*non-Xt command line options*/
  if (seed<=0) seed=getseed();   /*random number gen seed*/
  r250_init(seed);
  fprintf(stderr,"RNG seed %ld\n",seed);
  
  XtCallCallbacks(wmTemp,XmNactivateCallback,NULL);
  XtPopdown(XtParent(wTempMan));
  XtSetSensitive(wRunPushB,False); 
  /* Simulation cell size has to be adjusted if periodic bc in x */
/*  adjustsimarea(periodicXFlag,&(simArea.x),lParam);*/

    
  if (!readatoms){                   /*normal situation*/  
    XtCallCallbacks(wmNew,XmNactivateCallback,NULL);
    CalcAtoms();         /* calc the initial positions for atoms */
  } 
  else {                 /*load a saved situation in startup*/
    readatomsfromfile(); /* read the initial positions for atoms */
    CalcSimArea();
    InitRun();           /* initialize the system */
    XtSetSensitive(wRunPushB,True);
  }

  if (autostartFlag>0) { /* Batch run: start the simulation immediately */ 
    if (!readatoms)
      XtCallCallbacks(wPushBOK,XmNactivateCallback,NULL);
    XtCallCallbacks(wRunPushB,XmNactivateCallback,NULL);
  }
  
  Redraw();                       /* update the drawing window */
  XtAppMainLoop(app);     /*start the Xt application loop*/
  return 0;
}


/*------------------------------------------------------------*/
/*                 Signal HUP received.                       */
/*------------------------------------------------------------*/

void CatchHUP(int isig, siginfo_t *sinfo, void *dummy)
{
  fprintf(stderr,"%s: received signal %d from process %d.\n",
	  progname,isig,(int)sinfo->si_pid);
  sighupFlag=1;
}

/*------------------------------------------------------------*/
/*------------------------------------------------------------*/

/*
 * set initial values for simulation parameters
 */
void SetDefaults()
{
  /*FILE *pfd=NULL;*/
  /*char line[MAXLINE];
  char tfname[]="boundary.tscheme";
  int i;*/
  strcpy(tfname,"invalid file name");
  Pi = acos(-1.0);
  seed=3453545;

  
  epsilon = 1.02;
  epsilon2 = 1.0;
  lParam = 1.0;
  lParam2 = 0.96;
  relaxation = 0.0;
  cut = 2.1;
  Rminol=0.60;

  boundaryFlag = True;   
  verticalFlag = False;
  suggestFlag = False;  
  
  borderW = 1.0;
   
  crackFlag = False;         /* crack on/off                          */
  crack.angle = 30.0;       /* crack angle                           */ 
  crack.center.x = 0.0; /*center*/
  crack.center.y = 16.0;
  crack.size.x = 0.6;
  crack.size.y = 9.2;
  
  strcpy(tfname,"boundary.tscheme");
  tapplyFlag = False;
  tadjustFlag = True;
  tadjust2Flag = True;

  wWinSeamFlag = False;
  seam.disp.x = 0.0;
  seam.disp.y = 0.0;
  seam.Amove = 0.0;
  seam.Bmove = 0.0;
  seam.upcut = 0.0;
  seam.downcut = 0.0;
  seam.upcuta = 0.0;
  seam.downcuta = 0.0;
  seam.boundpos = 5.0;    
  seam.rotL = 0.0;
  seam.rotR = 0.0;

  periodicXFlag = False;
  
  deltaT = 0.002; 
  updateStep = /*(int)(50000*0.002)*/ 2;
  plotuStep = 4;
  stepCount = 0; 
  runningFlag = False;  
  initFlag = False;
  
  fixedFlag = False;
  strainFlag= False;
  strainStep = 0.0;
  strainP = -1.0;
  strainT = 800.0;
  pressure = 0.0;
  
  iTscheme = 0;
  cool.cooltemp = 50;
  cool.starttime = 10000;
  cool.tempvalue = 15;
  cool.slope = -0.001;
  cool.QuenchFlag = False;
  cool.TScaleFlag = False;
  cool.langevinFlag = False;
  cool.programon = 0;
  
  cgInit = 1;
  cgFlag = False;
  cgLineMinFlag = False;
  cgprevgrad = NULL;
  semisteepFlag = False;
  cgBracketFlag = False;
  cgMaxLMSteps = 5;
  cgGraphSteps = cgMaxLMSteps;
  cgSemiLMFlag = False;

  stopFlag=0;

  modarea = crack;      /*non-initial crack initial parameters*/
  modarea.angle = 0.0;       /* crack angle                           */ 
  modarea.center.x = 0.0; /*center*/
  modarea.center.y = 0.0;
  modarea.size.x = 1.0;
  modarea.size.y = 10.0;
  modvel.x = 0.0;
  modvel.y = 0.0;
  modinfo=M_MODRM;
  modallFlag = False;
  modrelatFlag = False;
  showselFlag = True;
  wWinModFlag = False;
  
  savetscheme = False;
  
  wWinnebFlag = False;
  nebk = 10.0;
  images = 20;
  cImage=0;
  grImage=0;
  nebip1 = 0.0;
  nebip2 = images-1;
  nebrelFlag=0;
  nebec = 0.0001;
  nebcc = 0.0001;
  nebcStop = False;
  wnebPlotSize.x = 300;
  wnebPlotSize.y = 100;
  nebcpi = 1.0;
  nebaddamount = 1.0;
  
  nebrepFlag=0;
  nebrepA=15.0;
  nebrepAlpha=5.0; 

  wWinTrackFlag = False;
  SetTrackFlag = False; 
  showTrackFlag = False;
  saveTrackFlag = False;
  PotFact = 0.88;
  RRc = 1.7;
  BFactor = 3.5;
  NumberofDefects=0;
  
  ani.aniFlag = False;  
  ani.limitsFlag = False; 
  ani.beginat = 1000.0;
  ani.endat = 50000.0;
  ani.waittime = 100.0;
  ani.snaphelp = 0.0;
  ani.endendFlag = False;
  wWinAniFlag = False;
  
  colorranges.Update = True;
  colorranges.hide  = False;
  colorranges.EkMin = 0.0;
  colorranges.EkMax = 33.0;
  colorranges.EpMin = -8.0;
  colorranges.EpMax = -1.0;
  colorranges.SMin = -60.0;
  colorranges.SMax = 60.0;
  colorranges.FMin = 0.0;
  colorranges.FMax = 20.0;

  sinfo = SKIN;
  avrCut = 0.0; 

  ae = NULL;
 // ra = NULL; 
  wDrawSize.x=700; wDrawSize.y=450;
  size.x = 100.0; size.y=60.0;
  zoom.y = 45.0;
  zoom.x = zoom.y *wDrawSize.x/wDrawSize.y;/*zoom.x can also be freely chosen*/
  center.x = 0.0; center.y = 0.0;
  edge.x = wDrawSize.x; edge.y=wDrawSize.y;
  diam = 0.96;
  plotdata.xscale = 2.0 / 10.0;
  plotdata.yscalingratio = 1.2;
  plotdata.rescale = 0;
  xstep = 2.0;
  printe=0; 
  strcpy(readdir,"");
  strcpy(savedir,"");

  /* Read some parameters from file (if the file exists).*/
  ROWParams("read","boundary.param");
  
#ifdef HARMONIC
  cut      = 1.3;
#endif
  if (verticalFlag && periodicXFlag) {
      XtVaSetValues(wPerxToggle, XmNset,False, NULL);
      XtCallCallbacks(wPerxToggle,XmNvalueChangedCallback,NULL);
  }
  if (savetscheme) {
    if (! (tsavefile = fopen ("save.tscheme", "w"))){
      fprintf(stderr,"Could not open file save.tscheme\n");
      bexit(-1);
    } 
  }
  c[IMAGESEL].max = images-1;
  c[NEB_IP1].max = images-1;
  c[NEB_IP2].max = images-1;
  c[IMAGECP].max = images-1;
  CalcSimArea();
}
/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/

void CalcSimArea()
{  
  if (periodicXFlag) 
    edge.x=0.0;
  else
    edge.x=wDrawSize.x;
  edge.y=wDrawSize.y;
  
  simArea.x  = (1.0 + edge.x / 100.0) * size.x;
  simArea.y  = (1.0 + edge.y / 100.0) * size.y;

  /* Simulation cell size has to be adjusted if periodic bc in x */
  adjustsimarea(periodicXFlag,&(simArea.x),lParam);
}

/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/

/*auxiliary function definitions for ROWParams*/
void rowd(char *varname,double *varptr,ROWPstruct s) 
{
  if (s.writingmode) {
    fprintf(s.pfile,varname);
    if (strlen(varname)<8) fprintf(s.pfile,"\t");
    fprintf(s.pfile,"\t%f\n",*varptr);
  }
  else {
    if (strcmp(varname,s.temp)==0) *varptr=s.dvar;
  }
}
void rowi(char *varname,int *varptr,ROWPstruct s) 
{
  if (s.writingmode) {
    fprintf(s.pfile,varname,"\t");
    if (strlen(varname)<8) fprintf(s.pfile,"\t");
    fprintf(s.pfile,"\t%d\n",*varptr);
  }
  else {
    if (strcmp(varname,s.temp)==0) *varptr=s.ivar;
  }
}


/* ReadOrWriteParams(..) reads or writes global variable values. */ 
void ROWParams(char *action, char *pfname) 
{
  char line[MAXLINE];
  int tmp=1;
  ROWPstruct s;   
  s.writingmode=0;
  s.pfile=NULL;
  if (strcmp(action,"write")==0) s.writingmode=1;
  if (s.writingmode) 
    s.pfile=fopen(pfname,"w");
  else 
    s.pfile=fopen(pfname,"r");
  
  if (s.pfile != NULL)  {
    if (s.writingmode) {
      fprintf(stderr,"Wrote parameters file \'%s\'.\n",pfname);
      fprintf(s.pfile,"#\n#Saved parameters.\n#\n\n");
      fprintf(s.pfile,"#syntax: <variable_name><space(s)><value><space(s)>"
	      "<possible comments>\n\n");  
    }    
    else  
      fprintf(stderr,"Reading some parameters from file \'%s\'.\n",pfname);
    
    while ((fgets(line,MAXLINE,s.pfile)!=NULL) || tmp>0)  {
      if (!s.writingmode) {
	sscanf(line,"%s %lf",s.temp,&s.dvar);
	sscanf(line,"%s %d",s.temp,&s.ivar);
      }
      if (s.writingmode) fprintf(s.pfile,"###Simulation area size\n");   
      rowd("size.x",&size.x,s);
      rowd("size.y",&size.y,s);
      rowi("wDrawSize.x",&wDrawSize.x,s);
      rowi("wDrawSize.y",&wDrawSize.y,s);
      
      
      if (s.writingmode) fprintf(s.pfile,"\n###Lattice params etc\n"); 
      rowd("lParam",&lParam,s);
      rowd("lParam2",&lParam2,s);
      rowd("epsilon",&epsilon,s);
      rowd("epsilon2",&epsilon2,s);
      rowd("cut",&cut,s);
      rowd("relaxation",&relaxation,s);
      rowd("deltaT",&deltaT,s); 
      rowd("Rminol",&Rminol,s); 
      
      if (s.writingmode) fprintf(s.pfile,"\n###Misc. parameters\n"); 
      rowi("autostartFlag",&autostartFlag,s);
      rowi("suggestFlag",&suggestFlag,s);
      rowi("periodicXFlag",&periodicXFlag,s);
      rowi("boundaryFlag",&boundaryFlag,s);
      rowi("verticalFlag",&verticalFlag,s);

      if (s.writingmode) fprintf(s.pfile,"\n###Modify area\n"); 
      rowi("modallFlag",&modallFlag,s);
      rowi("modrelatFlag",&modrelatFlag,s);
      rowi("showselFlag",&showselFlag,s);
        
      if (s.writingmode) fprintf(s.pfile,"\n###NEB parameters\n"); 
      rowi("images",&images,s);
      rowd("nebk",&nebk,s);
      rowi("nebcStop",&nebcStop,s);
      rowd("nebec",&nebec,s);
      rowd("nebcc",&nebcc,s);
      rowi("wnebPlotSize.x",&wnebPlotSize.x,s);
      rowi("wnebPlotSize.y",&wnebPlotSize.y,s);
      
      if (s.writingmode) fprintf(s.pfile,"\n###.events loading\n"); 
      rowi("tapplyFlag",&tapplyFlag,s);
      rowi("tadjustFlag",&tadjustFlag,s);
      rowi("tadjust2Flag",&tadjust2Flag,s);
      rowi("savetscheme",&savetscheme,s);

      if (s.writingmode) fprintf(s.pfile,"\n###Fit the boundary frame\n"); 
      rowd("seam.Amove",&seam.Amove,s);
      rowd("seam.Bmove",&seam.Bmove,s);
      rowd("seam.disp.x",&seam.disp.x,s);
      rowd("seam.disp.yy",&seam.disp.y,s);
      rowd("seam.dot",&seam.dot,s);
      rowd("seam.upcut",&seam.upcut,s);
      rowd("seam.upcuta",&seam.upcuta,s);
      rowd("seam.downcut",&seam.downcut,s);
      rowd("seam.downcuta",&seam.downcuta,s);
      rowd("seam.boundpos",&seam.boundpos,s);
      rowd("seam.rotL",&seam.rotL,s);
      rowd("seam.rotR", &seam.rotR,s);

      if (s.writingmode) fprintf(s.pfile,"\n###View controls\n"); 
      rowd("center.x",&center.x,s);
      rowd("center.y",&center.y,s);
      rowd("zoom.x",&zoom.x,s);
      rowd("zoom.y",&zoom.y,s);
      rowd("diam", &diam,s);
      rowi("updateStep",&updateStep,s);
      rowd("plotdata.yscalingratio",&plotdata.yscalingratio,s);
      rowi("plotuStep",&plotuStep,s);
      rowd("xstep",&xstep,s);

      if (s.writingmode) fprintf(s.pfile,"\n###Pressure and strain\n"); 
      rowd("pressure",&pressure,s);
      /*rowd("strainT",&strainT,s);
      rowd("strainP",&strainP,s);
      rowi("fixedFlag",&fixedFlag,s);
      rowd("borderW",&borderW,s);*/

      if (s.writingmode) fprintf(s.pfile,"\n###Temperature controls\n"); 
      rowd("cool.tempvalue",&cool.tempvalue,s);
      rowi("cool.TScaleFlag",&cool.TScaleFlag,s);
      rowi("cool.langevinFlag",&cool.langevinFlag,s);
      rowi("cool.QuenchFlag",&cool.QuenchFlag,s);
      rowd("cool.cooltemp",&cool.cooltemp,s);
      rowd("cool.slope",&cool.slope,s);
      rowd("cool.starttime",&cool.starttime,s);
      rowi("cool.programon",&cool.programon,s);
      rowi("cgFlag",&cgFlag,s);
      rowi("semisteepFlag",&semisteepFlag,s);
      rowi("cgBracketFlag",&cgBracketFlag,s);
      rowi("cgMaxLMSteps",&cgMaxLMSteps,s);
      rowi("cgLineMinFlag",&cgLineMinFlag,s);

      if (s.writingmode) fprintf(s.pfile,"\n###Animation controls\n"); 
      rowi("ani.aniFlag",&ani.aniFlag,s);
      rowd("ani.waittime",&ani.waittime,s);
      rowi("ani.limitsFlag",&ani.limitsFlag,s);
      rowd("ani.beginat",&ani.beginat,s);
      rowd("ani.endat",&ani.endat,s);
      rowi("ani.endendFlag",&ani.endendFlag,s);
      rowd("ani.snaphelp",&ani.snaphelp,s);
      
      if (s.writingmode) fprintf(s.pfile,"\n###Tracking controls\n"); 
      rowi("SetTrackFlag", &SetTrackFlag,s);
      rowi("saveTrackFlag",&saveTrackFlag,s);
      rowi("showTrackFlag",&showTrackFlag,s);
      rowd("RRc",&RRc,s);
      rowd("PotFact",&PotFact,s);
      rowd("BFactor",&BFactor,s);

      if (s.writingmode) fprintf(s.pfile,"\n###Color ranges\n"); 
      rowi("colorranges.Update",&colorranges.Update,s);
      rowd("colorranges.EkMin",&colorranges.EkMin,s);            
      rowd("colorranges.EkMax",&colorranges.EkMax,s); 
      rowd("colorranges.EpMin",&colorranges.EpMin,s);            
      rowd("colorranges.EpMax",&colorranges.EpMax,s);
      rowd("colorranges.SMin",&colorranges.SMin,s);            
      rowd("colorranges.SMax",&colorranges.SMax,s);
      rowd("colorranges.FMin",&colorranges.FMin,s);            
      rowd("colorranges.FMax",&colorranges.FMax,s);
      tmp=0;
    }
    
    fclose(s.pfile);
    c[IMAGESEL].max = images-1;
    c[NEB_IP1].max = images-1;
    c[NEB_IP2].max = images-1;
    c[IMAGECP].max = images-1;
    nebip2 = images-1;
  }
  else {
    fprintf(stderr,"Can\'t open file \'%s\'.\n",
	    pfname);
  }
}
 

/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/

/*reads command line arguments and does what has to be done*/
void getsomearguments(int argc, char **argv)
{
  int i,argerror;
    
  argerror=0;
  for (i=0;i<argc;i++) {
    if (strcmp(argv[i],"-clean")==0) {
      system("rm coords.????.i???.??? dump.????.xwd boundary.????.param Trk*.dat velocities.????.i???.dat 2> /dev/null");
      fprintf(stderr,"\nRemoved files coords.????.i???.??? dump.????.xwd boundary.????.param Trk*.dat velocities.????.i???.dat.\n");
      bexit(-1);
    } else if (strcmp(argv[i],"-start")==0) {
      autostartFlag=1;
    } else if (strcmp(argv[i],"-readatoms")==0) {
      sscanf(argv[++i],"%s",xyzloadnumber);
      readatoms=1;
      autostartFlag=0;
    } else if (strcmp(argv[i],"-dir")==0) {
      sscanf(argv[++i],"%s",readdir);
      strcpy(savedir,readdir);
    } else if (strcmp(argv[i],"-rdir")==0) {
      sscanf(argv[++i],"%s",readdir);
    } else if (strcmp(argv[i],"-sdir")==0) {
      sscanf(argv[++i],"%s",savedir);
    } else if (strcmp(argv[i],"-pe")==0) {
      sscanf(argv[++i],"%d",&printe);
    } else if (strcmp(argv[i],"-seed")==0) {
      sscanf(argv[++i],"%ld",&seed);
    } else {
      if (strstr(argv[i],"-")!=NULL) argerror=1;  /*unknown parameter*/
    }
  }
  if (argerror==1) {
    fprintf(stderr, "\nCommand line options for boundary:\n"
	    "-bg COLOR      set the background color for widgets to COLOR"
	    " (e.g. red)\n"
	    "-fg COLOR      set the widget foreground color to COLOR\n"
	    "-clean         remove all saved situation files (.xyz etc)\n"
	    "-pe INTEGER    print total kinetic energy and potential to "
	    "stdout at intervals\n               of INTEGER timesteps\n\n"
	    "-readatoms X   read atom positions from a coords.X.xyz file"
	    " where X is a 4-digit\n               number, e.g.\'0004\' \n"
	    "-dir STRING    set read & write directory for coordinate files\n"
	    "               example: -dir test/                        \n"
	    "-readdir STRING set read directory for coordinate files\n"
	    "-savedir STRING set save directory for coordinate files\n"
	    "\n-seed INTEGER  feed random number generator with INTEGER\n"
	    "-start         automatically start running the simulation \n");

    bexit(-1);
  }
}

/*---------------------------------------------------------*/
/*---------------------------------------------------------*/

/*get random number generator seed */

long int getseed()
{
  long int i;
  struct timeval tp;
  if (gettimeofday(&tp,(struct timezone *)NULL)==0) {  
    i=/* tp.tv_sec+ */tp.tv_usec;
    i=i%10000000|1;
    return i;
  } else {
    return -1;
  }
}



/*---------------------------------------------------------*/
/*---------------------------------------------------------*/




/* reads tscheme (in gui it's renamed to .events) from a file */

void readTscheme(char *tfname)
{
  char line[MAXLINE],tag[MAXLINE],modtype[MAXLINE];
  int i,j;
  double time,tmp,width,height;
  FILE *f;
  i=0;
  iTschemeMax=-1;
 
  if ((f=fopen(tfname,"r"))) {
    fprintf(stderr,"Reading .events file \'%s\'.\n",tfname); 
    for (i=0;i<3;i++) {
      fgets(line,MAXLINE,f); 
      /*fprintf(stderr,"  %s",line);*/
    }
    fgets(line,MAXLINE,f);
    sscanf(line,"%s %lf %lf", tag, &width, &height);
    /*fgets(line,MAXLINE,f);*/
    while (fgets(line,MAXLINE,f)!=NULL) {
      i++;
      sscanf(line,"%s %lf",tag,&time);
      j=0;
      while ((tag[j]=toupper(tag[j]))!='\0') j++;
      if (iTschemeMax>MAXTSCHEME) {
	fprintf(stderr,
		"Too many .events entries; should be no more than %d.\n",
		MAXTSCHEME);
	bexit(-1);
      }
      if (strcmp(tag,STschemeTscaleOn)==0) {
	iTschemeMax++;
	evTscheme[iTschemeMax]=TschemeTscaleOn;
	tTscheme[iTschemeMax]=time;
	sscanf(line,"%s %lf %lf",tag,&time,&tmp);
	TTscheme[iTschemeMax]=tmp;
      } 
      else if (strcmp(tag,STschemeTscaleOff)==0) {
	iTschemeMax++;
	evTscheme[iTschemeMax]=TschemeTscaleOff;
	tTscheme[iTschemeMax]=time;
	TTscheme[iTschemeMax]=VOIDTEMP;
      } 
      else if (strcmp(tag,STschemeRandomize)==0) {
	iTschemeMax++;
	evTscheme[iTschemeMax]=TschemeRandomize;
	tTscheme[iTschemeMax]=time;
	TTscheme[iTschemeMax]=VOIDTEMP;
      } 
      else if (strcmp(tag,STschemeQuenchOn)==0) {
	iTschemeMax++;
	evTscheme[iTschemeMax]=TschemeQuenchOn;
	tTscheme[iTschemeMax]=time;
	TTscheme[iTschemeMax]=VOIDTEMP;
      } 
      else if (strcmp(tag,STschemeQuenchOff)==0) {
	iTschemeMax++;
	evTscheme[iTschemeMax]=TschemeQuenchOff;
	tTscheme[iTschemeMax]=time;
	TTscheme[iTschemeMax]=VOIDTEMP;
      } 
      else if (strcmp(tag,STschemeQuit)==0) {
	iTschemeMax++;
	evTscheme[iTschemeMax]=TschemeQuit;
	tTscheme[iTschemeMax]=time;
	TTscheme[iTschemeMax]=VOIDTEMP;
      } 
      else if (strcmp(tag,STschemeModifyArea)==0) {
	iTschemeMax++;
	evTscheme[iTschemeMax]=TschemeModifyArea;
	tTscheme[iTschemeMax]=time;
	
	sscanf(line,"%s %lf %s %lf %lf %lf %lf %lf %lf %lf",tag,&time,
	       modtype,
	       &ModXYT[iTschemeMax].x, &ModXYT[iTschemeMax].y,
	       &ModAT[iTschemeMax].center.x, &ModAT[iTschemeMax].center.y,
	       &ModAT[iTschemeMax].size.x, &ModAT[iTschemeMax].size.y,
	       &ModAT[iTschemeMax].angle);
	if (tadjustFlag) {
	  ModAT[iTschemeMax].center.x *= size.x/width;
	  ModAT[iTschemeMax].center.y *= size.y/height;
	}
	if (tadjust2Flag) {
	  ModAT[iTschemeMax].size.x *= size.x/width;
	  ModAT[iTschemeMax].size.y *= size.y/height;
	}
	if (strcmp(modtype,"remove")==0)     TTscheme[iTschemeMax]=M_MODRM;
	else if (strcmp(modtype,"stop")==0)  TTscheme[iTschemeMax]=M_MODSTOP;
	else if (strcmp(modtype,"type1")==0) TTscheme[iTschemeMax]=M_MODTYPE1;
	else if (strcmp(modtype,"type2")==0) TTscheme[iTschemeMax]=M_MODTYPE2;
	else if (strcmp(modtype,"vel")==0)   TTscheme[iTschemeMax]=M_MODVEL;
	else if (strcmp(modtype,"fix")==0)   TTscheme[iTschemeMax]=M_MODFIX;
	else if (strcmp(modtype,"force")==0) TTscheme[iTschemeMax]=M_MODFORCE;
	else if (strcmp(modtype,"rmfix")==0) TTscheme[iTschemeMax]=M_MODRMF;
	else if (strcmp(modtype,"fixx")==0)  TTscheme[iTschemeMax]=M_MODFIXX;
	else if (strcmp(modtype,"fixy")==0)  TTscheme[iTschemeMax]=M_MODFIXY;
	else if (strcmp(modtype,"tag")==0)   TTscheme[iTschemeMax]=M_MODTAG;
	else if (strcmp(modtype,"move")==0)  TTscheme[iTschemeMax]=M_MODMOVE;
	else if (strcmp(modtype,"rot")==0)   TTscheme[iTschemeMax]=M_MODROT;
      }
    }
    iTschemeMax++;
    tTscheme[iTschemeMax]=LARGETIME;
    
    fclose(f);
  } 
  else {
    iTschemeMax=-1;
    fprintf(stderr,".events file \'%s\' not found.\n",tfname);
  }
}

/*---------------------------------------------------------*/
/*---------------------------------------------------------*/

/*void UpdateEventsInfo() {
  char buff[250], line[250];
  int i;
  sprintf(buff, "");
  
  for (i=0; i < iTschemeMax; i++) {
    sprintf(line,"%s %f\n",TschemeNames[evTscheme[i]],tTscheme[i]);
    strcat(buff, line);

  }
  XmTextSetString(wTextEvents, buff);
}*/

/*-------------------------------------------------------------*/
/*-------------------------------------------------------------*/


/* applies tscheme, checks which actions to do in the current time step */

void checkTscheme()
{
  double t;
  int oldmodinfo;
  char buff[MAXLINE];
  t=stepCount*1000.0*deltaT;
  while (iTscheme<iTschemeMax && t>=tTscheme[iTscheme]) {
    switch (evTscheme[iTscheme]) {
    case TschemeTscaleOn:
      cool.TScaleFlag = True;
      XtVaSetValues(wTScaleToggle,XmNset,cool.TScaleFlag,NULL); 
      *c[TEM].value=TTscheme[iTscheme];
      sprintf(buff,"%g",TTscheme[iTscheme]);
      XmTextSetString(c[TEM].text,buff);
      break;
    case TschemeTscaleOff:
      cool.TScaleFlag = False;
      XtVaSetValues(wTScaleToggle,XmNset,cool.TScaleFlag,NULL); 
      break;
    case TschemeRandomize:
      RandomizeVelocities();
      break;
    case TschemeQuenchOn:
      cool.QuenchFlag = True;
      XtVaSetValues(wQuenchToggle,XmNset,cool.QuenchFlag,NULL); 
      break;
    case TschemeQuenchOff:
      cool.QuenchFlag = False;
      XtVaSetValues(wQuenchToggle,XmNset,cool.QuenchFlag,NULL); 
      break;
    case TschemeQuit:
      SnapShot();
      fprintf(stderr,"Stop at %g fs (%ld steps).\n",t,stepCount);
      bexit(0);
      break;
    case TschemeModifyArea:
      oldmodinfo = modinfo;
      modinfo=(int)TTscheme[iTscheme];
      modvel.x=ModXYT[iTscheme].x;
      modvel.y=ModXYT[iTscheme].y;
      ModifyArea(ModAT[iTscheme]);
      modinfo=oldmodinfo;
      break;
    case TschemeUnknown:
    default:
      fprintf(stderr,"Invalid T scheme event at %g: %d.\n",
	      tTscheme[iTscheme],evTscheme[iTscheme]);
      break;
    }
    iTscheme++;
  }
  
}



/*---------------------------------------------------------*/
/*---------------------------------------------------------*/

void LoadImageFromFile(int image)
{
  static char line[MAXLINE];
  static char typestring[2];
  int i,t,atag,fixvel,atom;
  double x,y,z, cforcex, cforcey;
  FILE *file, *file2;
  static char fname[MAXLINE]="coords.";
  static char fdname[MAXLINE]="adata.";
  sprintf(fname,"%scoords.%s.i%03d.xyz",readdir, xyzloadnumber,image);
  
  
  //fprintf(stderr,"LoadImageFromFile %d\n",image);
  //fprintf(stderr,"nimet: %s   %s", fname, f2name);
  if (file=fopen(fname,"r")) {
    fprintf(stderr,"Reading configuration from file \'%s\'.\n",fname);
    fscanf(file,"%d\n",&nAtom);
    fgets(line,MAXLINE,file); 
    
    for (i=nAtom*image;i<nAtom*(image+1);i++) {
      fgets(line,MAXLINE,file);
      sscanf(line,"%s %lg %lg %lg",typestring,&x,&y,&z);
      if (strncmp("SI",typestring,2)==0) t=1;
      else if (strncmp("GE",typestring,2)==0) t=2;
      else if (strncmp("GA",typestring,2)==0) t=-1;
      else if (strncmp("AS",typestring,2)==0) t=-2;
      /*line[0]=' '; line[1]=' ';
	sscanf(line," %lg %lg %lg",&x,&y,&z);*/
      ae[i].fixvel=0;
      ae[i].atag=0;
      ae[i].cforce.x=0.0; 
      ae[i].cforce.y=0.0;
      ae[i].atype=t;
      ae[i].r.x=x;
      ae[i].r.y=y;
      ae[i].rv.x=0.0;
      ae[i].rv.y=0.0;
    }
    fclose(file);
  } else {
    fprintf(stderr,"LoadImageFromFile: Can\'t read file \'%s\'.\n",fname);
    return;
    //bexit(-1);
  }

  sprintf(fdname,"%sadata.%s.i%03d.dat",readdir,xyzloadnumber,image);
  if (file2=fopen(fdname,"r")) {
    //fprintf(stderr,"Reading file \'%s\'.\n",fdname);
    //fgets(line,MAXLINE,file2); 
    while (!feof(file2)) {
      fgets(line,MAXLINE,file2);
      //fprintf(stderr,"%s\n",line);
      if (line[1]!='#') {
	sscanf(line,"%d %d %lg %lg %d",&atom, &fixvel, &cforcex, &cforcey, &atag);
	ae[atom].fixvel = fixvel;
	ae[atom].atag = atag;
	ae[atom].cforce.x = cforcex; 
	ae[atom].cforce.y = cforcey;
      }
    }
    fclose(file2);
  } else {
    //fprintf(stderr,"Can\'t read file \'%s\'.\n",fdname);
  }

  fprintf(stderr,"Finished reading \'%s\'.\n",fname);

}



/*---------------------------------------------------------*/
/*---------------------------------------------------------*/


/* Loads a saved situation. Reads atom coordinates and system parameters 
   and immediately starts the simulation (check main()).*/

void readatomsfromfile()
{
  FILE *fda,*file2;
  char afname[MAXLINE]="coords.",pfname[MAXLINE]="boundary.",line[MAXLINE];
  char af2name[MAXLINE]="adata.",typestring[2];
  int i,t,tag,fixv,atom;
  double x,y,z,cfx,cfy;
  sprintf(afname,"%scoords.",readdir);
  sprintf(pfname,"%sboundary.",readdir);
  sprintf(af2name,"%sadata.",readdir);
 
  strcat(afname,xyzloadnumber);
  strcat(afname,".i000");
  strcat(afname,".xyz");
  strcat(pfname,xyzloadnumber);
  strcat(pfname,".param");

  strcat(af2name,xyzloadnumber);
  strcat(af2name,".i000");
  strcat(af2name,".dat");


  /*afname=strcat(afname,afname);*/
  if ((fda=fopen(afname,"r"))) {
    fprintf(stderr,"Reading configuration from file \'%s\'.\n",afname);
    ROWParams("read",pfname);
  } else {
    fprintf(stderr,"readatomsfromfile: Can\'t read file \'%s\'.\n",afname);
    bexit(-1);
  }


  fscanf(fda,"%d\n",&nAtom);
  /*
  if ((ae = (AtomD*)malloc(images*(nAtom) * sizeof(AtomD))) == NULL) {
      printf("Error allocating memory for ae!\n");
      bexit(-1);
  }*/
  ALLOC(ae, images* (nAtom), AtomD);
  if (ae == NULL) {
    printf("Error allocating memory for ae in control.c!\n");
    bexit(-1);
  }
  
  fprintf(stderr,"Number of atoms %d\n",nAtom);
  fgets(line,MAXLINE,fda); 
  fprintf(stderr,"%s\n",line);

  fclose(fda);
  LoadImageFromFile(0);

  /* check for border atoms */
  for (i = 0; i < nAtom; i++) {
    ae[i].borderatom=0;
    if (ae[i].r.x < -size.x/2.0+borderW) ae[i].borderatom |= BORDER_LEFT;
    if (ae[i].r.x >  size.x/2.0-borderW) ae[i].borderatom |= BORDER_RIGHT;
    if (ae[i].r.y < -size.y/2.0+borderW) ae[i].borderatom |= BORDER_BOTTOM;
    if (ae[i].r.y >  size.y/2.0-borderW) ae[i].borderatom |= BORDER_TOP;
  }
  
  //fclose(fda);

  /*if ((file2=fopen(af2name,"r"))) {
    fprintf(stderr,"Reading data from file \'%s\'.\n",af2name);
    fgets(line,MAXLINE,file2); 
    for (i=0;i<nAtom;i++) {
      fgets(line,MAXLINE,file2);
      sscanf(line,"%d %lg %lg %d",&fixv, &cfx, &cfy, &tag);
      ae[i].fixvel=fixv;
      ae[i].cforce.x=cfx;
      ae[i].cforce.y=cfy;
      ae[i].atag=tag;
    }
    fclose(file2);
  } else {
    fprintf(stderr,"Can\'t read file \'%s\'.\n",af2name);
  }*/
  
  return;
}

/*---------------------------------------------------------*/
/*---------------------------------------------------------*/

    
/*Rounds a double to nearest integral double. This is often a library 
  function in math.h but not on all machines so it is included here.*/

double nint(double x) 
{
  if ((ceil(x)-x)<=(x-floor(x)))
    return ceil(x);
  else
    return floor(x);
}

/*---------------------------------------------------------*/
/*---------------------------------------------------------*/

/* exits the program and closes possible open .events file */
void bexit(int value) {
  if (savetscheme)
    fclose(tsavefile);
  exit(value);
}

/*---------------------------------------------------------*/
/*---------------------------------------------------------*/

/* copies atom coordinates and other relevant info for initial
   neb setup*/
void nebCopy() {
  if (images > 1) {
    int i,j; 
    for (i=1;i<images;i++) {
      for (j=0;j<nAtom;j++) {
	ae[nAtom*i+j]=ae[j];
      }
    }
  }
  return;
}

