#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>  /*DBL_MAX*/ 
#include <float.h>   /*DBL_MAX*/

/* time */
#include <sys/types.h>
#include <time.h>
#include <sys/timeb.h>

#include "control.h"
#include "graphics.h"
#include "simu.h"
#include "fitsigma.h"
#include "calc.h"
#include "callback.h"
#include "gvar.h"    /*load global variables*/
#include "track.h"
#include "r250.h"
#include "conjgrad.h"

/*Definitions used in simu.c.*/

/* default values */
#define SMALLTEMP 1.0e-15
#define MAXTYPE 3

#ifdef ONEBODY
/* One-body energies and forces */
double OneBody(double x, double y, double *e, double *fx, double *fy);
#endif

double Energy;
double eps[MAXTYPE][MAXTYPE] = {{0,0,0},{0,0,0},{0,0,0}};
double cutoffdist[MAXTYPE][MAXTYPE]={{0,0,0},{0,0,0},{0,0,0}};
double sig[MAXTYPE][MAXTYPE]={{0,0,0},{0,0,0},{0,0,0}};   
#ifdef HARMONIC
double r0[MAXTYPE][MAXTYPE]={{0,0,0},{0,0,0},{0,0,0}};   
double kharm[MAXTYPE][MAXTYPE]={{0,0,0},{0,0,0},{0,0,0}};   
#endif
double temperature;
int *bottom, *top;           /* arrays for fixed atoms */
int borderNMax;            /* size of bottom[] and top[] */
//MtrxD *StressT = NULL;  /*Stress Tensor...*/
int strainCount = 0;



extern Container c[];
//extern XtWorkProcId id;

/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/


/*
 * initialize computation
 */
void InitRun()
{
  int i;
#ifdef HARMONIC
  double r26;
  double ktmp;
#endif
  double lpar;
  double rCut,rCutB,rCutAB;/* time step and potential cutoff distances */
  double  sigma,sigma2;    /* potential normalizing parameters */
  kinEnergy = 1.0;
  /* defaults */
  temperature =TEMPERATURE;
  

  /* Cut off distances in different cases AA BB AB and BA */
  rCut = cut *lParam / sqrt(2.0);
  rCutB = cut *lParam2/ sqrt(2.0);
  rCutAB = pow(((SQ4(rCut)*SQ(rCut)+SQ4(rCutB)*SQ(rCutB))/2),(1.0/6.0));
  cutoffdist[1][1]= rCut;
  cutoffdist[2][2]=rCutB;
  cutoffdist[1][2]=  rCutAB;
  cutoffdist[2][1]= cutoffdist[1][2];
  

  /*Potential parameters*/

  
  sigma =  pow(0.50*(pow(cut,-7.0)-1.0)/(pow(cut,-13.0)-1.0),
	       1.0/6.0)*lParam/sqrt(2.0);
  sigma2 = pow(0.50*(pow(cut,-7.0)-1.0)/(pow(cut,-13.0)-1.0),
	       1.0/6.0)*lParam2/sqrt(2.0);
  /*fprintf(stderr,"lParam %.12g %.12g\n",lParam,lParam2);
  fprintf(stderr,"sigma  %.12g %.12g\n",sigma,sigma2);*/
  sigma=fitsigma(lParam/sqrt(2.0),rCut);
  sigma2=fitsigma(lParam2/sqrt(2.0),rCutB);
  sigma  *= 1.0/sqrt(2.0)/0.6999;  /* For some reason sigma has to be 
                                      scaled up a bit!!!!! */
  sigma2 *= 1.0/sqrt(2.0)/0.6999;  /* For some reason sigma has to be 
                                      scaled up a bit!!!!! */
  /*fprintf(stderr,"sigma  %.12g %.12g\n",sigma,sigma2);*/
  sig[1][2]=pow(((SQ4(sigma)*SQ(sigma)+SQ4(sigma2)*SQ(sigma2))/2),(1.0/6.0));
  sig[2][1] = sig[1][2];
  sig[1][1]=sigma;
  sig[2][2]=sigma2;
  
  eps[1][1] = epsilon;
  eps[2][2] = epsilon2;
  eps[1][2]=(2*sqrt(epsilon*epsilon2)*SQ(sigma)*sigma*sigma2*SQ(sigma2))/
  ((SQ4(sigma)*SQ(sigma)+SQ4(sigma2)*SQ(sigma2)));
  eps[2][1] = eps[1][2];

#ifdef HARMONIC
  /* Equilibrium position (used in harmonic potential) */
  r26=pow(2.0,1.0/6.0);
  r0[1][1]=r26*sig[1][1];
  r0[1][2]=r26*sig[1][2];
  r0[2][1]=r26*sig[2][1];
  r0[2][2]=r26*sig[2][2];
  /* Harmonic potential force constants */
  ktmp=9.0/pow(2.0,1.0/3.0);
  kharm[1][1]=ktmp*UNITENERGY*eps[1][1]/pow(sig[1][1],2.0);
  kharm[1][2]=ktmp*UNITENERGY*eps[1][2]/pow(sig[1][2],2.0);
  kharm[2][1]=ktmp*UNITENERGY*eps[2][1]/pow(sig[2][1],2.0);
  kharm[2][2]=ktmp*UNITENERGY*eps[2][2]/pow(sig[2][2],2.0);
#endif
  
  nAtomOrig = nAtom;
  if (rCut>=rCutB){
    VSCopy(cells, 1.0 / rCut, simArea);
  }
  else{
    VSCopy(cells, 1.0 / rCutB, simArea);
  }
  if (avrCut > 0.099)
    VSCopy(avCells, 1.0 / avrCut, simArea);
  else {
    avCells.x = 0;
    avCells.y = 0;
  }
  
  /* memory (re)allocation */
  /*ALLOC(rv, images * nAtom, VecD);
  ALLOC(ra, images * nAtom, VecD);
  ALLOC(cforce, images * nAtom,VecD);*/
  ALLOC(cellList, /*#images * #*/nAtom + VProd(cells), int);
  ALLOC(avv, VProd(avCells), VecD);
  ALLOC(cgprevgrad, nAtom, VecD);
  
  /*ALLOC(nebPotEnergy,images,double);
  ALLOC(nebkink, images,double);
  ALLOC(nebfix, images,int);*/
  ALLOC(neb,images,nebImage);
  for (i=0; i < images; i++) {
	 neb[i].fix=0;
	 neb[i].pot=0.0;
	 neb[i].kink=0.0;
  }
  neb[0].fix=1;
  neb[images-1].fix=1;
  /*border atom max count calculations*/
  if (lParam<lParam2)
    lpar=lParam;
  else
    lpar=lParam2;
  borderNMax = (int) ceil(3*2*size.x/(lpar/sqrt(2.0)));
  ALLOC(bottom, borderNMax, int);
  ALLOC(top, borderNMax, int);
  
  /*ALLOC(potEn,images*nAtomOrig,double);
  ALLOC(stressij,images*nAtomOrig,double);*/

  /*---Stress--*/
  /*ALLOC(StressT,images*nAtom,MtrxD);*/

  /* determine maximum and minimum coordinates of the atoms */

  xMin = 0.0;
  xMax = 0.0;
  yMin = 0.0;
  yMax = 0.0;
  
  for (i = 0; i < nAtom; i++) {
    /*cforce[i].x=0.0;cforce[i].y=0.0;*/
    if (ae[i].r.x < xMin)      xMin = ae[i].r.x;
    else if (ae[i].r.x > xMax) xMax = ae[i].r.x;
    else if (ae[i].r.y < yMin) yMin = ae[i].r.y;
    else if (ae[i].r.y > yMax) yMax = ae[i].r.y;
  }

  /* initialize computation */
  InitialState();
  ComputeForces();
  potEnergy /= nAtom;
  CalcEnergy();

  /* set flags */
  /*changeTemp = 0;*/
  stepCount = 0;
  strainFlag = 0;
  iTscheme=0;
  strainCount = 0;
  initFlag = True;

  /* test */
  Redraw();
  UpdatePlot(1);
}

/*---------------------------------------------------*/
/*---------------------------------------------------*/


/* one neb cycle */
void nebCycle() {

  if (nebrelFlag) {    /* compute one image in the normal way */
    cImage=grImage;
    if (cgFlag) {
      cgMinimizer();
      potEnergy /= nAtom;
    } else {
      ComputeForces();
      /*if (nebrepFlag)
	NebRepulsion();*/
      neb[gtype(grImage)].pot = potEnergy; 
      potEnergy /= nAtom;
      LeapfrogIntegrate();
      CalcEnergy();
    }
  } else {       /* use the neb algorithm */

    /* calculate forces and potential energies in each image */
    nebecFlag = 1; nebccFlag=1;
    for (cImage=0; cImage < images; cImage++) {
      if (neb[cImage].fix < 2) {
	ComputeForces();
	if ((fabs(potEnergy)>nebMINTANG) && 
	    (fabs((neb[cImage].pot-potEnergy)/potEnergy)>nebec)) nebecFlag=0;
	neb[cImage].pot=potEnergy;
	if (neb[cImage].fix == 1) neb[cImage].fix = 2;
      }

    }
    
    potEnergy = neb[(int)grImage].pot / nAtom;
        
    /* update forces and velocities */
    for (cImage=1; cImage < (images-1); cImage++) {
      if (0 == neb[cImage].fix)
	nebUpdate(); 
    }
    /* update positions */ 
    for (cImage=1; cImage < (images-1); cImage++) {
      if (0 == neb[cImage].fix)
	NebUpdatePositions(); 
    }
    cImage = grImage;
    CalcEnergy();    
  }
}

/*---------------------------------------------------*/
/*---------------------------------------------------*/




/*
 * work procedure: single md step
 */

Boolean MdCycle()
{
  int i,updstep;

  double newtemp,currenttime;

  //fprintf(stderr,"MdCycle %d %d\n",stepCount,printe); 

  /*if (progFlag==0) return False;*not in use?*/  
  if (nAtom == 0) return False;
  ++stepCount;
  if (tapplyFlag) checkTscheme();
  
  if (nAtom == 0) return False;
  
  if (wWinnebFlag)
    nebCycle();
  else {
    if (cgFlag) {
      cgMinimizer();
      potEnergy /= nAtom;
    }
    else {
      ComputeForces();
      potEnergy /= nAtom;
      LeapfrogIntegrate();
      CalcEnergy();
    }
  }
 
  /* Temperature */ 
  
  currenttime=deltaT*stepCount*1000.0;

  if (!cgFlag) {
    if (cool.programon) {
      if (currenttime<cool.starttime) 
	SetNewTemp(cool.cooltemp/temperature);
      else {
	newtemp=cool.cooltemp + (cool.slope*(currenttime-cool.starttime));
	SetNewTemp(newtemp/temperature);
      }
    } else {
      if (cool.TScaleFlag) SetNewTemp(cool.tempvalue/temperature);
    }
  }



  if (stepCount % plotuStep == 0)  
    UpdatePlot(0);
  
  updstep=(int)((double)updateStep/(1000.0*deltaT)+0.5);
  if (updstep<1) updstep=1;
  if (stepCount % updstep == 0 ) {
    Redraw();
    //RedrawPlot(0);
    if (wWinnebFlag && !nebrelFlag) RedrawnebPlot();
  }
  
  if (sighupFlag) {
    SnapShot();
    sighupFlag=0;
  }

  if (printe>0) {
    if (stepCount % printe == 0) {
      fprintf(stdout,"%.15g %.15g %.15g %d %g\n",
	      (double)stepCount*1000.0*deltaT,kinEnergy,potEnergy,stepCount,
	      (double)clock()/CLOCKS_PER_SEC);
      fflush(stdout);
    }
  }

  if (ani.aniFlag && ani.snaphelp<(stepCount * 1000 * deltaT)){
    if (!ani.limitsFlag){
      SnapShot();
      ani.snaphelp += ani.waittime;
    } else {
      if (ani.snaphelp<ani.beginat){
	ani.snaphelp = ani.beginat;
      } else if ((stepCount * 1000 * deltaT-ani.waittime/10)<=ani.endat) {
	SnapShot();
	ani.snaphelp += ani.waittime;
      } else if (!ani.endendFlag) {
	ani.aniFlag = False;
      } else {
	bexit(0);
      }
    }
  }
  
  /*------------------Adding Tracking-----------------*/
  if (SetTrackFlag){
    Trackprocedure();
    if (saveTrackFlag) 
      savingprogram();
  };
  /*---------------------------------------------------*/

  /* stop simu if neb has converged */
   if (wWinnebFlag && nebcStop && nebccFlag && nebecFlag && runningFlag) {
     XtCallCallbacks(wRunPushB,XmNactivateCallback,(XtPointer) RUN);
     fprintf(stdout,"ec: %d, cc: %d\n",nebecFlag,nebccFlag);
     Redraw();
     UpdatePlot(0);
     RedrawnebPlot();
     return True;
   }
   
   /* stops the run */
   if (stopFlag) {
     stopFlag=0;
     XtCallCallbacks(wRunPushB,XmNactivateCallback,(XtPointer) RUN);
     fprintf(stderr,"boundary stopped\n");
     Redraw();
     UpdatePlot(0);
     return True;
   }

  return False;
}


/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/



/*
 * compute force and potential energy using cell method. The heart of the 
 * program.
 */
/* convenience macros */

#define DO_ATOM for (n = 0; n < nAtom; n ++)
#define DO_CELL(j, m) for (j = cellList[m]; j >= 0; j = cellList[j])

void ComputeForces()
{
  VecI cc, m1v, m2v, iof[] = {{0,0}, {1,0}, {1,1}, {0,1}, {-1,1}};
  VecD dr, invWid, rs;
  double ri2[MAXTYPE][MAXTYPE]={{0,0,0},{0,0,0},{0,0,0}};
  double ri6[MAXTYPE][MAXTYPE]={{0,0,0},{0,0,0},{0,0,0}};
  double urc[MAXTYPE][MAXTYPE],durc[MAXTYPE][MAXTYPE];
  double fcVal, r2,sigma6,rang2,rang6;
  int j1, j2, offset, btot, ttot,i,j;
  int help1,help2;
  int avenei,navenei;
 
  VecI cellmin,cellmax;
  double dtmp;
#ifdef HARMONIC
  double rtmp;
#endif
  double stress;
  unsigned int c, m1, m2, n;

#ifdef ONEBODY
  /* For one-body energies and forces */
  double e1,f1x,f1y;
#endif

  /* Langevin */
  double disipvalue=0.8;
  double Tempki;

  //fprintf(stderr,"*** sinfo %d (%d%d%d)\n",sinfo,SXX,SXY,SYY);

  /* This loop calculates Lennart-Jones potential at cutoff distance (urc) */
  /* and then the derivative (durc) of potential, in every possible case.  */
  for(i=0;i<MAXTYPE;i++){
    for(j=0;j<MAXTYPE;j++){
      dtmp=sig[i][j];
      sigma6 = dtmp*dtmp*dtmp*dtmp*dtmp*dtmp;
      if (SQ(cutoffdist[i][j])!=0){
	ri2[i][j]=1.0/SQ(cutoffdist[i][j]);
	ri6[i][j]= ri2[i][j] * ri2[i][j] * ri2[i][j];
	urc[i][j] = UNITENERGY*eps[i][j]*sigma6*ri6[i][j]
	  *(sigma6*ri6[i][j]-1.0);
	durc[i][j] = 12.0*UNITENERGY*eps[i][j]
	  *sigma6*ri6[i][j]*(0.5-sigma6*ri6[i][j])/cutoffdist[i][j];  
      }
    }
  }
  

  simArea2.x=simArea.x/2.0;
  simArea2.y=simArea.y/2.0;
  btot = 0;
  ttot = 0;
  
  /* initialize linked lists */

  for (n = nAtomOrig; n < nAtomOrig + VProd(cells); n++) cellList[n] = -1;
  VDiv(invWid, cells, simArea);

  /* assign atoms to cells */

  for (n = 0+cImage*nAtom; n < nAtom*(cImage+1); n++) {
    ae[n].potEn = ae[n].stressij = 0.0;
    /* remove atoms that get outside the cell area */
    if (fabs(ae[n].r.x) > 0.5 * simArea.x 
	|| fabs(ae[n].r.y) > 0.5 * simArea.y) {
      ae[n]=ae[--nAtom];
      fprintf(stderr,"Warning: removing atoms.\n");
      stopFlag=1;
    } 
    else { /* atom inside cell area */
      VSAdd(rs, ae[n].r, 0.5, simArea);
      /* assign atom to a simulation cell */
      VMul(cc, rs, invWid);
      c = VLinear(cc, cells) + nAtomOrig;
      cellList[(int) (n-cImage*nAtom)] = cellList[c];
      cellList[c] = n-cImage*nAtom;
    }
  }

  /* Cell limits depend on boundary conditions */
  if (periodicXFlag) {
    cellmin.x=-1;
    cellmax.x=cells.x+1;
  } else {
    cellmin.x=0;
    cellmax.x=cells.x;
  }    
  cellmin.y=0;
  cellmax.y=cells.y;
      
  avenei=0;
  navenei=0;

    /* zero accelerations or set them to defaut values */
  //DO_ATOM {
  for (n = 0+cImage*nAtom; n < nAtom*(cImage+1); n ++) {
    
    navenei++;

    ae[n].ra.x = ae[n].cforce.x*1.0;
    ae[n].ra.y = ae[n].cforce.y*1.0;
    
        
    VZero(ae[n].StressT.x);
    VZero(ae[n].StressT.y);
  }
  potEnergy = 0.0;
  /* loop over all cells */
  for (m1v.y = 0; m1v.y < cells.y; m1v.y++) {
    for (m1v.x = 0; m1v.x < cells.x; m1v.x++) {
      m1 = m1v.y * cells.x + m1v.x + nAtomOrig;
      /* loop over offsets between adjacent cells */
      for (offset = 0; offset < 5; offset ++) {
	VAdd(m2v, m1v, iof[offset]);
	/* check that the cell inside the simulation area */
	if (m2v.x>=cellmin.x && m2v.x<cellmax.x 
	    && m2v.y>=cellmin.y && m2v.y<cellmax.y) {
	  /* Handle periodicity */
	  if (periodicXFlag) {
	    if (m2v.x<0) m2v.x=cells.x-1;
	    if (m2v.x>=cells.x) m2v.x=0;
	  }
	  
	  m2 = VLinear(m2v, cells) + nAtomOrig;
	  /* loops over atoms in each cell */
	  DO_CELL (j1, m1) {
	    DO_CELL (j2, m2) {
	      /* ensure each atom pair examined only once */
	      if (m1 != m2 || j2 < j1) {
		VSub(dr, ae[(int)(j1+nAtom*cImage)].r, 
		     ae[(int)(j2+nAtom*cImage)].r);		
		if (periodicXFlag) {
		  if (dr.x<-simArea2.x) dr.x+=simArea.x;
		  if (dr.x>=simArea2.x) dr.x-=simArea.x;
		}
		
		r2 = VLenSq(dr);
		/* fprintf(stderr,"%g\n",sqrt(r2)); */
		/* if atoms within interaction range */
		help1=(int)fabs((double)ae[j1].atype);
		help2=(int)fabs((double)ae[j2].atype);
		if (r2 <SQ(cutoffdist[help1][help2])) {
		  avenei++;
#ifdef HARMONIC
		  rtmp=r0[help1][help2]-sqrt(r2);
		  fcVal=2.0*kharm[help1][help2]*rtmp/sqrt(r2);
		  Energy=kharm[help1][help2]*rtmp*rtmp;
#else
		  rang2 = 1.0 / r2;
		  rang6 = rang2 * rang2 * rang2;
		  dtmp=sig[help1][help2];
		  dtmp=dtmp*dtmp*dtmp*dtmp*dtmp*dtmp;
		  
		  fcVal = 12.0*UNITENERGY*eps[help1][help2]*dtmp
		    *rang6*(dtmp*rang6-0.5)*rang2
		    +durc[help1][help2]*sqrt(rang2);  
		  
		  /* update accelerations and energy sum */
		  Energy = UNITENERGY*eps[help1][help2]*dtmp*rang6
		    *(dtmp*rang6-1.0)+
		    (cutoffdist[help1][help2]-sqrt(r2))*
		    durc[help1][help2]-urc[help1][help2];
#endif		  
		  VVSAdd(ae[(int)(j1+nAtom*cImage)].ra, fcVal, dr);
		  VVSAdd(ae[(int)(j2+nAtom*cImage)].ra, - fcVal, dr);
		  
		  /*if (sinfo == SXX || sinfo == SXY || sinfo == SYY) {
		    if (showxx||showxy||showyy) {*/
		  if (sinfo == SXX) stress=fcVal*dr.x*dr.x;
		  else if (sinfo == SXY) stress=fcVal*dr.x*dr.y;
		  else if (sinfo == SYY) stress=fcVal*dr.y*dr.y;
		  else stress=0.0;
		  ae[(int)(j1+nAtom*cImage)].stressij+=stress/2.0;
		  ae[(int)(j2+nAtom*cImage)].stressij+=stress/2.0;
		  //} 
		  ae[(int)(j1+nAtom*cImage)].potEn+=Energy/2.0;
		  ae[(int)(j2+nAtom*cImage)].potEn+=Energy/2.0;
		  potEnergy +=Energy;
		  
		}
	      }
	    } /* DO_CELL (j2, m2) */
	    /*test m1...*/
	    if (pressure != 0.0 && m1 == m2) { /* search edge atoms */
	      if (ae[(int)(j1+nAtom*cImage)].r.y < yMin + borderW 
		  && ae[(int)(j1+nAtom*cImage)].r.y < 0 
		  /*&& btot < borderNMax*/) {
		bottom[btot++] = j1;
	      } 
	      else if (ae[(int)(j1+nAtom*cImage)].r.y > yMax - borderW 
		       && ae[(int)(j1+nAtom*cImage)].r.y > 0 
		       /*&& ttot < borderNMax*/) {
		top[ttot++] = j1;
 	      }
	    } /* if (m1 == m2 */
	  } /* DO_CELL (j1, m1) */
	} /* if (m2v.x >= 0 ... */
      } /* for (offset = 0 ... */
    } /* for (m1x = 0 ... */
  } /* for (m1y = 0 ... */

  /* if (navenei>0) fprintf(stderr,"avenei %g\n",2.0*(double)avenei/(double)navenei); */
#ifdef ONEBODY
  /* Add one-body energy and force */
  for (n = 0; n < nAtom; n++) {
    OneBody(r[n].x,r[n].y,&e1,&f1x,&f1y);
    ae[n].potEn+=e1;
    potEnergy+=e1;
    ae[n].ra.x+=f1x;
    ae[n].ra.y+=f1y;
  }
#endif


  /* Apply pressure */ 
  if (pressure != 0.0 && btot != 0 && ttot != 0) {
    fcVal = pressure * (xMax - xMin) / btot;
    for (j1 = 0; j1 < btot; ++j1) ae[bottom[j1]].ra.y += fcVal;
    fcVal = pressure * (xMax - xMin) / ttot;
    for (j1 = 0; j1 < ttot; ++j1) ae[top[j1]].ra.y -= fcVal;
  }
  
  potEnmin=1.0e20;
  potEnmax=-1.0e20;
   if (sinfo == SXX || sinfo == SXY || sinfo == SYY) {
     /*if (showxx||showxy||showyy) {*/
    stressijmin=1.0e20;
    stressijmax=-1.0e20;
  }

   /* calculate max and min potential energy */
  for (n = 0+cImage*nAtom; n < nAtom*(cImage+1); n++) {
    if (ae[n].potEn < potEnmin) potEnmin=ae[n].potEn;
    if (ae[n].potEn > potEnmax) potEnmax=ae[n].potEn;

     /*------Test for include temp control (Langevin and Berensen)-----*/
    /* A suitable value for disipvalue is 0.8 */

    if (cool.langevinFlag && !cgFlag){
      //fprintf(stderr,"langevin\n");
      Tempki=UNITTEMP * VLenSq(ae[n].rv)/(2.0);
      ae[n].ra.x += - disipvalue*ae[n].rv.x + 1.0*sqrt(disipvalue)
	*sqrt(2.0*cool.tempvalue/(UNITTEMP*deltaT))*(gasdev(&seed));
      ae[n].ra.y += - disipvalue*ae[n].rv.y + 1.0*sqrt(disipvalue)
	*sqrt(2.0*cool.tempvalue/(UNITTEMP*deltaT))*(gasdev(&seed));
    }


    if (sinfo == SXX || sinfo == SXY || sinfo == SYY) {
    /*if (showxx||showxy||showyy) {*/
      if (ae[n].stressij < stressijmin) stressijmin=ae[n].stressij;
      if (ae[n].stressij > stressijmax) stressijmax=ae[n].stressij;
    }
  }
  
  if (wWinnebFlag && nebrepFlag)
    NebRepulsion();


  for (n=cImage*nAtom; n < (cImage+1)*nAtom; n++) {
     if (ae[n].fixvel & FIXX) ae[n].ra.x = 0.0;
     if (ae[n].fixvel & FIXY) ae[n].ra.y = 0.0;
  }
  
}

/*-------------------------------------------------------------*/
/*-------------------------------------------------------------*/


void NebRepulsion()
{
  int n;
  double dx,dy,tmp;
  VecD coord0;
  VecD coord;
  //fprintf(stderr,"A:%g\n",nebrepA);
  //fprintf(stderr,"Alpha: %g\n",nebrepAlpha);
  for (n=cImage*nAtom; n < nAtom*(cImage+1); n++) {
    if (ae[n].atag == 1) {
      coord0=ae[n-cImage*nAtom].r;
      coord=ae[n].r;
      dx=coord.x-coord0.x;
      dy=coord.y-coord0.y;
      
      tmp=nebrepA*exp(-nebrepAlpha*(SQ(dx)+SQ(dy)));
      ae[n].potEn += tmp;
      ae[n].ra.x += dx*tmp*nebrepAlpha*2;
      ae[n].ra.y += dy*tmp*nebrepAlpha*2;
      potEnergy += tmp;
    }
  }
}



/*-------------------------------------------------------------*/
/*-------------------------------------------------------------*/

/*
 * Averaging vectors: calculate avarage of rArg
 */
void Average(AtomD *rArg)
{
  int i, c;
  VecD invWid, rs;
  VecI cc;
  if ((avCells.x != 0) && (avCells.y != 0)) {
  /* zero number of atoms in cells */
  for (i = 0; i < avCells.x * avCells.y; i++) {
    avn[i] = 0;
    avr[i].x = 0.0;
    avr[i].y = 0.0;
    avv[i].x = 0.0;
    avv[i].y = 0.0;
    avatype[i]=0;
  }
  VDiv(invWid, avCells, simArea);
 
  /* find the cells of the atoms and calculate net value of argument */
  for (i = 0+cImage*nAtom; i < nAtom*(cImage+1); i++) {
    VSAdd(rs, ae[i].r, 0.5, simArea);
    VMul(cc, rs, invWid);
    c = VLinear(cc, avCells);
    avr[c].x += ae[i].r.x;
    avr[c].y += ae[i].r.y;
    avv[c].x += rArg[i].rv.x;
    avv[c].y += rArg[i].rv.y;
    avn[c]++;
    if (ae[i].atype==1 || ae[i].atype==-1){
      avatype[c]-=1;
    }
    else
      avatype[c]+=1;
  }
  /* divide net values by the number of atoms inside cells */
  for (i = 0; i < avCells.x * avCells.y; i++) {
    if (avn[i] > 1) {
      avr[i].x /= (double)avn[i];
      avr[i].y /= (double)avn[i];
      avv[i].x /= (double)avn[i];
      avv[i].y /= (double)avn[i];
    }
    else if (avn[i]==0)
      avr[i].x = DBL_MAX;
  }
  }
}

/*-------------------------------------------------------------*/
/*-------------------------------------------------------------*/


/* calculates the distance between two neb images,this and the previous one */
double nebdi(int image) 
{
  double len=0.0;
  double dlen;
  int i;
  if (image > 0) {
    for(i=image*nAtom; i < (image+1)*nAtom; i++) {
      dlen = (ae[i].r.y - ae[i-nAtom].r.y);
      len += dlen*dlen;
      dlen = (ae[i].r.x - ae[i-nAtom].r.x);
      PBCdx(dlen);
      len += dlen*dlen;
    }
    return sqrt(len);
  } else {
    return 0;
  }
}


/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/

/* auxiliary function */
VecD nebTang(int image, int n) 
{
  VecD tang;
  tang.y = (ae[n+image*nAtom].r.y - ae[n+(image-1)*nAtom].r.y);
  tang.x = (ae[n+image*nAtom].r.x - ae[n+(image-1)*nAtom].r.x);
  PBCdx(tang.x);
  return tang;
}

/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/


/* neb main function: calculate spring forces, nudging and velocity
   projection */
void nebUpdate() 
{
  int n;
  double nebttl; /* neb tangent total length */
  double adott;     /* accel. dot neb tang */
  //VecD tang1[nAtom],tang2[nAtom]; /* full neb tangent */
  VecD t1,t2;    /* 2 components of the neb tangent */
  double t1l, t2l;
  double t1_dot_t2; /* t1 dot t2 */
  VecD nebac;   /* neb accel. correction (springs and nudging) */
  double Vplus, Vmed, Vminus, DVmax, DVmin, DV1, DV2; /* p. energies for neb */
  double sabs, sabs1, sabs2; /* absolute value of total spring force */
  double nebal;   /* length of corrected accel. */
  //VecD nebua;   /* corrected unit acceleration */
  double nebvdota;
  double dtmp;
  
  VecD *tang;
  static VecD *tang1=NULL; /* full neb tangent */
  static VecD *tang2=NULL;
  static VecD *spring=NULL;

  if (tang1 == NULL) {
    if ((tang1 = (VecD*)malloc(nAtom * sizeof(VecD))) == NULL) {
      printf("Error allocating memory for tang1!\n");
      bexit(-1);
    }
  }

  if (tang2 == NULL) {
    if ((tang2 = (VecD*)malloc(nAtom * sizeof(VecD))) == NULL) {
      printf("Error allocating memory for tang2!\n");
      bexit(-1);
    }
  }

  if (spring == NULL) {
    if ((spring = (VecD*)malloc(nAtom * sizeof(VecD))) == NULL) {
      printf("Error allocating memory for spring!\n");
      bexit(-1);
    }
  }

  /* calculate tangents and kinks */
  t1l=0.0; t2l=0.0; t1_dot_t2 = 0.0;
  for (n=0; n<nAtom; n++) {
    t1 = nebTang(cImage+1,n);
    t2 = nebTang(cImage,n);
    tang1[n] = t1; 
    tang2[n] = t2;
    
    t1l += t1.x*t1.x + t1.y*t1.y;
    t2l += t2.x*t2.x + t2.y*t2.y;
    t1_dot_t2 += t1.x * t2.x + t1.y * t2.y;
    
  }
  t1l = sqrt(t1l);
  if (t1l < nebMINTANG) t1l = nebMINTANG;
  t2l = sqrt(t2l);
  if (t2l < nebMINTANG) t2l = nebMINTANG;
  neb[cImage].kink = t1_dot_t2 / (t1l * t2l);

  /* neb tangent estimate */
  nebttl = 0.0;
  Vplus = neb[cImage+1].pot;
  Vmed = neb[cImage].pot;
  Vminus = neb[cImage-1].pot;
  
  if ((Vplus > Vmed) && (Vmed > Vminus)) {
    tang = tang1;
    nebttl = t1l;
  } else if ((Vplus < Vmed) && (Vmed < Vminus)) {
    tang = tang2;
    nebttl = t2l;
  } else { /* an extremum on the potential curve */
    tang = tang1;
    DV1 = fabs(Vplus-Vmed); DV2 = fabs(Vminus-Vmed);
    if (DV1 > DV2) {
      DVmax = DV1;
      DVmin = DV2;
    } else {
      DVmax = DV2;
      DVmin = DV1;
    }
    
    if (Vplus > Vminus) {
      for (n=0; n<nAtom; n++) {
	tang[n].x = tang1[n].x*DVmax + tang2[n].x*DVmin;
	tang[n].y = tang1[n].y*DVmax + tang2[n].y*DVmin;
	nebttl += tang[n].x * tang[n].x + tang[n].y * tang[n].y;
      }
    } else {
      for (n=0; n<nAtom; n++) {
	tang[n].x = tang1[n].x*DVmin + tang2[n].x*DVmax;
	tang[n].y = tang1[n].y*DVmin + tang2[n].y*DVmax;
	nebttl += tang[n].x * tang[n].x + tang[n].y * tang[n].y;
      } 
    }
    nebttl = sqrt(nebttl);
  }

  if (nebttl < nebMINTANG) nebttl = nebMINTANG;
  
  /* absolute value of spring force and value of ra dot tang */
  sabs = 0.0; sabs1 = 0.0; sabs2 = 0.0;
  adott = 0.0;
  for (n = cImage*nAtom; n < nAtom*(cImage+1); n++) {
    
    /* normalize the tangent */
    tang[n-cImage*nAtom].x /= nebttl;
    tang[n-cImage*nAtom].y /= nebttl;

    /* spring force */
    dtmp=ae[n+nAtom].r.x-ae[n].r.x;
    PBCdx(dtmp);
    sabs1 += dtmp*dtmp;
    sabs1 += (ae[n+nAtom].r.y-ae[n].r.y)*(ae[n+nAtom].r.y-ae[n].r.y);
    dtmp=ae[n].r.x-ae[n-nAtom].r.x;
    PBCdx(dtmp);
    sabs2 += dtmp*dtmp;
    sabs2 += (ae[n].r.y-ae[n-nAtom].r.y)*(ae[n].r.y-ae[n-nAtom].r.y);
    
    /*ra dot tang */
    adott += ae[n].ra.x * tang[n-cImage*nAtom].x;  
    adott += ae[n].ra.y * tang[n-cImage*nAtom].y;
  }
  sabs1 = sqrt(sabs1); 
  sabs2 = sqrt(sabs2);
  sabs = nebk*(sabs1-sabs2);
  
  
  /* apply neb nudging and calc. length of new accel. */
  nebal=0.0;
  for (n = cImage*nAtom; n < (cImage+1)*nAtom; n++) {
    nebac.x = tang[n-cImage*nAtom].x * (sabs - adott);
    nebac.y = tang[n-cImage*nAtom].y * (sabs - adott);
    
    ae[n].ra.x += nebac.x;
    ae[n].ra.y += nebac.y;
    nebal += ae[n].ra.x * ae[n].ra.x;
    nebal += ae[n].ra.y * ae[n].ra.y;
  }    
  //nebal=sqrt(nebal);
  if (nebal < nebMINTANG) nebal = nebMINTANG;
  
  nebvdota = 0.0;
  for (n = cImage*nAtom; n < nAtom*(cImage+1); n++) {
    nebvdota += (ae[n].rv.x * ae[n].ra.x + ae[n].rv.y * ae[n].ra.y);
  }

  /* velocity Verlet projection, update velocities */
  for (n = cImage*nAtom; n < nAtom*(cImage+1); n++) {
    if (nebvdota < 0) {
      if (!(ae[n].fixvel & FIXX)) ae[n].rv.x = deltaT*ae[n].ra.x;
      if (!(ae[n].fixvel & FIXY)) ae[n].rv.y = deltaT*ae[n].ra.y;
    } else {
      if (!(ae[n].fixvel & FIXX)) 
	ae[n].rv.x = nebvdota * ae[n].ra.x / nebal + deltaT*ae[n].ra.x;
      if (!(ae[n].fixvel & FIXY)) 
	ae[n].rv.y = nebvdota * ae[n].ra.y / nebal + deltaT*ae[n].ra.y;
    }    
  }
}

/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/

void NebUpdatePositions() {
  VecD prevr;
  int n;
  double ddx;
  for (n = cImage*nAtom; n < nAtom*(cImage+1); n++) {
    prevr=ae[n].r;
    
    ae[n].r.y += deltaT * ae[n].rv.y;
    ae[n].r.x += deltaT * ae[n].rv.x;
    PBCx(ae[n].r.x);
    
    ddx=prevr.x - ae[n].r.x;
    PBCdx(ddx);
    if ((fabs(ae[n].r.x) > nebMINTANG) 
	&& (fabs((ddx) / ae[n].r.x) > nebcc))
      nebccFlag = 0;
    if ((fabs(ae[n].r.y) > nebMINTANG) 
	&& (fabs((prevr.y - ae[n].r.y) / ae[n].r.y) > nebcc))
      nebccFlag = 0;
  }
}

/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/


/*
 * integration over single timestep
 */
void LeapfrogIntegrate()
{
  int n;
  double vdota,len2;
  static double friction=0.9;
  //VecD prevr;

  yMin = 0.0;
  yMax = 0.0;
  
  for (n = cImage*nAtom; n < nAtom*(cImage+1); n++) {
    
    vdota = 0.0;
    
    /* alternative minimization, test version */
    /*if (minFlag) {
      len2 = ae[n].ra.x * ae[n].ra.x + ae[n].ra.y * ae[n].ra.y;
      vdota = ae[n].rv.x * ae[n].ra.x + ae[n].rv.y * ae[n].ra.y;
      if (1==1) {
	if (len2 < SMALLDOUBLE) len2 = SMALLDOUBLE;
	if (vdota<0) len2 *= 2.0;
	if (!(ae[n].fixvel & FIXX)) 
	  ae[n].rv.x = vdota * ae[n].ra.x / len2;
	if (!(ae[n].fixvel & FIXY)) 
	  ae[n].rv.y = vdota * ae[n].ra.y / len2;
      } else {
	if (!(ae[n].fixvel & FIXX)) ae[n].rv.x = 0.0;
	if (!(ae[n].fixvel & FIXY)) ae[n].rv.y = 0.0;
      }
      
    }*/

    if (cool.QuenchFlag && (cImage == grImage)) 
      vdota = ae[n].rv.x * ae[n].ra.x + ae[n].rv.y * ae[n].ra.y;
    

    /* vertical boundary */
    if (verticalFlag) {
      
      if (cool.QuenchFlag) {
	if (vdota<0.0 && !(ae[n].fixvel & FIXY)) ae[n].rv.y*=friction;
      }
      if (!(ae[n].fixvel & FIXY)) ae[n].rv.y += deltaT * ae[n].ra.y;
      //r[n].y += deltaT * rv[n].y;
      
      if (!fixedFlag || ae[n].atype>0) {
	if (cool.QuenchFlag && vdota<0.0) {
	  if (!(ae[n].fixvel & FIXX)) ae[n].rv.x*=friction;
	}
	if (!(ae[n].fixvel & FIXX)) ae[n].rv.x += deltaT * ae[n].ra.x;
	//r[n].x += deltaT * rv[n].x;
      } else {
	if (!(ae[n].fixvel & FIXX)) ae[n].rv.x = 0.0; 
      }
      
    } else { /* horizontal boundary */
      
      if (cool.QuenchFlag) {
	if (vdota<0.0) {if (!(ae[n].fixvel & FIXX)) ae[n].rv.x*=friction;}
      }
      if (!(ae[n].fixvel & FIXX)) ae[n].rv.x += deltaT * ae[n].ra.x;
      //r[n].x += deltaT * rv[n].x;
      
      if (!fixedFlag || ae[n].atype>0) {
	if (vdota<0.0) {if (!(ae[n].fixvel & FIXY)) ae[n].rv.y*=friction;}
	if (!(ae[n].fixvel & FIXY)) ae[n].rv.y += deltaT * ae[n].ra.y;
	//r[n].y += deltaT * rv[n].y;
      } else {
	if (!(ae[n].fixvel & FIXY)) ae[n].rv.y = 0.0; 
      }
                  
    }

    ae[n].r.y += deltaT * ae[n].rv.y;
    ae[n].r.x += deltaT * ae[n].rv.x;

    if (periodicXFlag) {
      if (ae[n].r.x < -simArea2.x) ae[n].r.x += simArea.x;
      if (ae[n].r.x >= simArea2.x) ae[n].r.x -= simArea.x;
    }
    
    /* find the y-edges of the system */
    if (ae[n].r.y < yMin)
      yMin = ae[n].r.y;
    else if (ae[n].r.y > yMax)
      yMax = ae[n].r.y;
  }
}  


/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/

  
/* change temperature */
/* We take the square root of the scaling factor here. */

void SetNewTemp(double f)
{
  double vvSum;
  int n;
  

  if (nAtom==0) return;

  if (f<0.0) 
    f=0.0;
  else 
    f=sqrt(f);
  
  vvSum = 0.;

  for (n=grImage*nAtom; n < (grImage+1)*nAtom; n++) {
    if (ae[n].fixvel==0) VScale(ae[n].rv, f);
    vvSum += VLenSq(ae[n].rv);
  }

  temperature = UNITTEMP*vvSum/(2.0*nAtom);
}


/*--------------------------------------------------------*/
/*--------------------------------------------------------*/

/*this is called when the user presses randomize button*/

void RandomizeVelocities()
{
  int n;
  double std;
  VecD vSum;

  /* fprintf(stderr,"RandomizeVelocities: %ld\n",seed); */
  VZero(vSum);
  std=sqrt(temperature/UNITTEMP);
  for (n=grImage*nAtom; n < (grImage+1)*nAtom; n++) {
  //DO_ATOM {
    if (ae[n].fixvel==0) {
      ae[n].rv.x=std*gasdev();
      ae[n].rv.y=std*gasdev();
    }
    VAdd(vSum, vSum, ae[n].rv);
  }
}


/*--------------------------------------------------------*/
/*--------------------------------------------------------*/

/*
 * initialize atom velocities
 */
void InitialState()
{
  VecD vSum;
  double ang, vMag;
  int n;
  
  stepCount = 0;

  vMag = sqrt(2. * temperature);
  VZero(vSum);
  DO_ATOM {
    /* velocity has fixed magnitude and random direction */
    ang = 2. * M_PI * randnum();
    /*atag[n]=0;
    fixvel[n]=0;*/
    ae[n].rv.x = cos(ang);
    ae[n].rv.y = sin(ang);
    VScale(ae[n].rv, vMag);
    VAdd(vSum, vSum, ae[n].rv);
    
  }
  /* ensure zero center of mass velocity */
  DO_ATOM {
    if (ae[n].fixvel==0)
      VVSAdd(ae[n].rv, - 1. / nAtom, vSum);
  }
}


/*-------------------------------------------------------------*/
/*-------------------------------------------------------------*/


/* Random number generator */
/* We now use r250!!       */

double randnum() {
  double r;
  r=dr250();
  return r;
}



/*-------------------------------------------------------------*/
/*-------------------------------------------------------------*/

double gasdev()
{
  static int iset=0;
  static double gset;
  double fac,rsq,v1,v2;
  
  if  (iset == 0) {
    do {
      v1=2.0*randnum()-1.0;
      v2=2.0*randnum()-1.0;
      rsq=v1*v1+v2*v2;
    } while (rsq >= 1.0 || rsq == 0.0);
    fac=sqrt(-2.0*log(rsq)/rsq);
    gset=v1*fac;
    iset=1;
    return v2*fac;
  } else {
    iset=0;
    return gset;
  }
}


/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/


/*
 * compute system average kinetic and potential energy and temperature
 */
void CalcEnergy()
{
  VecD v;
  double vvSum,vsq,accel;
  int n;
  
  kinEnmin=1.0e20;
  kinEnmax=-1.0e20;
  forcemin=1.0e20;
  forcemax=-1.0e20;
  vvSum = 0.0;
  //DO_ATOM {
  for (n = 0+nAtom*cImage; n < nAtom*(cImage+1); n ++) {
    if (fabs(ae[n].r.x) <= 0.5 * simArea.x 
	&& fabs(ae[n].r.y) <= 0.5 * simArea.y) {
      /* allow for half timestep */
      VSAdd(v, ae[n].rv, -0.5 * deltaT, ae[n].ra);
      vsq=VLenSq(v);
      vvSum += vsq;
      accel=sqrt(VLenSq(ae[n].ra));
      if (vsq<kinEnmin) kinEnmin=vsq;
      if (vsq>kinEnmax) kinEnmax=vsq;
      if (accel<forcemin) forcemin=accel;
      if (accel>forcemax) forcemax=accel;
    }
  }
  
  if (nAtom > 0) {
    kinEnergy = 0.5 * vvSum / nAtom;  
    temperature = UNITTEMP * kinEnergy;
    //potEnergy /= nAtom;
    kinEnmin*=0.5;
    kinEnmax*=0.5;
  }
  else {
    kinEnergy = 0.0;
    //potEnergy = 0.0;
    kinEnmin=-1.0;
    kinEnmax=1.0;    
  }
}



#ifdef ONEBODY
/* One-body energy and force */
double OneBody(double x, double y, double *e, double *fx, double *fy) 
{
  double sx,cx,sy,cy;
  static double E=5.0,a=0.2,b=0.2;
  sx=sin(a*x);
  cx=cos(a*x);
  sy=sin(b*y);
  cy=cos(b*y);
  *e=E*sx*cy;
  *fx=-E*a*cx*cy;
  *fy=E*b*sx*sy;
}
#endif


/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/

/* Remove overlapping atoms, i.e. if atoms are nearer to each */
/* other than Rminol one of the atoms is removed.             */

void RmOverlappingAtoms(double grImage, double Rminol, RArea area)
{
  int i,j,k,*rma;
  double dx,dy,dr2,rm2;

  fprintf(stderr,"Now in RmOverlappingAtoms: %g %g\n",grImage,Rminol);

  if ((rma=(int*)malloc(nAtom*sizeof(int)))==NULL) {
    fprintf(stderr,"Could not allocate rma in RmOverlappingAtoms.\n");
    exit(-1);
  }
  for (i=0;i<nAtom;i++) rma[i]=0;

  rm2=Rminol*Rminol;
  k=-1;

  for (i=grImage*nAtom; i<nAtom*(grImage+1); i++) {
    k++;
    if (Selected(i,area)) {
      for (j=grImage*nAtom; j<nAtom*(grImage+1); j++) {
	if (i!=j) {
	  dx=ae[i].r.x-ae[j].r.x;
	  dy=ae[i].r.y-ae[j].r.y;
	  if (periodicXFlag) {
	    if (dx<-simArea2.x) dx+=simArea.x;
	    if (dx>=simArea2.x) dx-=simArea.x;
	  }
	  dr2=dx*dx+dy*dy;
	  if (dr2<rm2) rma[k]=1;
	}
      }
    }
  }

  /* for (i=0;i<nAtom;i++) fprintf(stderr,"%1d",rma[i]);fprintf(stderr,"\n"); */
  k=-1;
  for (i=grImage*nAtom; i<nAtom*(grImage+1); i++) {
    k++;
    if (rma[k]==1) {
      ae[i]=ae[((int)grImage)*nAtom+nAtom-1];
      rma[k]=rma[nAtom];
      k--;
      i--;
      nAtom--;
    }
  }

  free(rma);
}  
